# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command list_notification_channels."""

import os
import socket
from unittest import mock

from django.conf import settings
from django.test import TestCase, override_settings

from debusine.django.management.tests import call_command


class InfoTests(TestCase):
    """Tests for the info command."""

    def test_info_defaults(self) -> None:
        """Info output with no overrides."""
        fqdn = settings.DEBUSINE_FQDN
        deb_fqdn = settings.DEBUSINE_DEBIAN_ARCHIVE_FQDN

        with (
            mock.patch(
                "shutil.get_terminal_size",
                return_value=os.terminal_size((65536, 25)),
            ),
            mock.patch(
                # We make sure DEBUSINE_FQDN is an FQDN, for tests
                # Return our value, in case we added an example domain.
                "socket.getfqdn",
                return_value=fqdn,
            ),
        ):
            stdout, stderr, exit_code = call_command("info")

        self.assertIn(f" * DEBUSINE_FQDN = {fqdn!r}", stdout)
        self.assertIn(
            "The current value matches the autodetected default"
            " from socket.getfqdn().",
            stdout,
        )

        # The test runner adds 'testserver' therefore editing the setting from
        # its autodetected value
        self.assertIn(
            f" * ALLOWED_HOSTS = [{fqdn!r}, {deb_fqdn!r}, 'localhost', "
            f"'127.0.0.1', 'testserver']",
            stdout,
        )
        self.assertIn(
            "The current value is changed from default"
            f" [{fqdn!r}, 'localhost', '127.0.0.1']",
            stdout,
        )

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_info_default_allowed_hosts(self) -> None:
        """Info output with ALLOWED_HOSTS having a default value."""
        fqdn = settings.DEBUSINE_FQDN
        with (
            mock.patch(
                "shutil.get_terminal_size",
                return_value=os.terminal_size((65536, 25)),
            ),
            mock.patch(
                # We make sure DEBUSINE_FQDN is an FQDN, for tests
                # Return our value, in case we added an example domain.
                "socket.getfqdn",
                return_value=fqdn,
            ),
            override_settings(ALLOWED_HOSTS=[fqdn, "localhost", "127.0.0.1"]),
        ):
            stdout, stderr, exit_code = call_command("info")

        self.assertIn(f" * DEBUSINE_FQDN = {fqdn!r}", stdout)
        self.assertIn(
            "The current value matches the autodetected default"
            " from socket.getfqdn().",
            stdout,
        )

        self.assertIn(
            f" * ALLOWED_HOSTS = [{fqdn!r}, 'localhost', '127.0.0.1']", stdout
        )
        self.assertIn(
            "The current value matches the autodetected default from"
            " DEBUSINE_FQDN.",
            stdout,
        )

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

    def test_info_changed_debusine_fqdn(self) -> None:
        """Info output with overridden DEBUSINE_FQDN."""
        with (
            mock.patch(
                "shutil.get_terminal_size",
                return_value=os.terminal_size((65536, 25)),
            ),
            override_settings(
                DEBUSINE_FQDN="test.example.org",
                # ALLOWED_HOSTS is built from DEBUSINE_FQDN
                ALLOWED_HOSTS=["foo", "testserver"],
            ),
        ):
            stdout, stderr, exit_code = call_command("info")

        fqdn = socket.getfqdn()
        self.assertIn(" * DEBUSINE_FQDN = 'test.example.org'", stdout)
        self.assertIn(
            f"The current value is changed from the default {fqdn!r}",
            stdout,
        )
        self.assertIn(" * ALLOWED_HOSTS = ['foo', 'testserver']", stdout)
        self.assertIn(
            "The current value is changed from default"
            " ['test.example.org', 'localhost', '127.0.0.1']",
            stdout,
        )

        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common utility methods."""

import logging
import socket
import subprocess
from functools import cached_property
from typing import Any

import yaml

logger = logging.getLogger(__name__)


class Configuration:
    """Configuration setup for the integration tests."""

    DEBUSINE_SERVER_USER = 'debusine-server'

    @classmethod
    def get_base_url(cls) -> str:
        """Return the base URL for the test server."""
        return f"https://{socket.getfqdn()}"


class RunResult:
    """Encapsulate the result of executing a command."""

    def __init__(
        self, stdout: str, stderr: str, returncode: int, cmd: list[str]
    ) -> None:
        """Initialize RunResult."""
        self.stdout = stdout
        self.stderr = stderr
        self.returncode = returncode
        self.cmd = cmd

    @cached_property
    def parsed_stdout(self) -> Any:
        """Return the command stdout parsed as YAML."""
        # Parse it only once and only if trying to use __getitem__.
        # Some cmd run by RunResult might not return valid Yaml
        # (so the yaml should only be parsed if it's used)
        try:
            return yaml.safe_load(self.stdout)
        except yaml.YAMLError as exc:
            raise ValueError(
                f'Error parsing\n'
                f"{self.cmd=}\n"
                f"{self.stdout=}\n"
                f"{self.stderr=}\n"
                f"{self.returncode=}\n"
                f"{exc=}"
            )

    def __getitem__(self, key: str) -> Any:
        """Return item from parsed stdout from the command."""
        if self.parsed_stdout is None:
            raise KeyError(
                f'KeyError: "{key}" cannot be found: stdout is None\n'
                f"{self.cmd=}\n"
                f"{self.stdout=}\n"
                f"{self.stderr=}\n"
                f"{self.returncode=}"
            )
        elif key not in self.parsed_stdout:
            raise KeyError(
                f'KeyError: "{key}" does not exist.\n'
                f"{self.cmd=}\n"
                f"{self.stdout=}\n"
                f"{self.parsed_stdout=}\n"
                f"{self.stderr=}\n"
                f"{self.returncode=}"
            )

        return self.parsed_stdout[key]


def run_as(user: str, cmd: list[str], *, stdin: str | None = None) -> RunResult:
    """Run cmd as user using sudo."""
    cmd = ['sudo', '-u', user, *cmd]

    return run(cmd, stdin=stdin)


def run(
    cmd: list[str], *, stdin: str | None = None, timeout: float | None = None
) -> RunResult:
    """Run cmd with stdin, return subprocess.CompletedProcess."""
    logger.debug('Exec: %s (stdin: %s)', cmd, stdin)

    completed_process = subprocess.run(
        cmd,
        text=True,
        capture_output=True,
        input=stdin,
        timeout=timeout,
    )

    return RunResult(
        completed_process.stdout,
        completed_process.stderr,
        completed_process.returncode,
        cmd,
    )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Signing-side no-operation task."""

from pathlib import Path
from typing import override

from debusine.artifacts.models import WorkRequestResults
from debusine.signing.tasks import BaseSigningTask
from debusine.signing.tasks.models import SigningNoopData


class SigningNoop(BaseSigningTask[SigningNoopData]):
    """Task that runs on server-side Celery workers and returns a boolean."""

    TASK_NAME = "noop"
    TASK_VERSION = 1

    @override
    def get_subject(self) -> str | None:
        return None

    @override
    def run(self, execute_directory: Path) -> WorkRequestResults:  # noqa: U100
        """Do nothing."""
        return self.data.result



# Generated by Django 4.2.21 on 2025-10-30 15:53

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0021_workrequest_completed_by'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='workflow_root',
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.DO_NOTHING,
                related_name='descendants',
                to='db.workrequest',
            ),
        ),
    ]



# Copyright 2020-2023 Enrico Zini <enrico@debian.org>
# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test of the backend for signon authentication using external providers."""

import re
from collections.abc import Collection
from typing import Any, override
from unittest import mock

import django.contrib.sessions.backends.base
from django.contrib import auth
from django.contrib.auth import get_user_model
from django.contrib.auth.models import AnonymousUser
from django.contrib.messages.storage.fallback import FallbackStorage
from django.core.exceptions import ImproperlyConfigured, PermissionDenied
from django.test import RequestFactory, override_settings

from debusine.db.models import Identity, User
from debusine.server.signon import providers
from debusine.server.signon.auth import SignonAuthBackend
from debusine.server.signon.middleware import RequestSignonProtocol
from debusine.server.signon.signon import MapIdentityFailed, Signon
from debusine.test.django import TestCase


class MockSession(django.contrib.sessions.backends.base.SessionBase):
    """In-memory session with no persistence, used for tests."""

    # TODO: is there a more standard way to provide a mock session?
    # Would SessionStore be enough?

    # @override
    # def exists(self, session_key: str) -> bool:
    #     """Mock SessionBase API endpoint."""
    #     return True

    @override
    def create(self) -> None:
        """Mock SessionBase API endpoint."""

    # @override
    # def save(self, must_create: bool = False) -> None:
    #     """Mock SessionBase API endpoint."""
    #     pass

    @override
    def delete(self, *args: Any) -> None:
        """Mock SessionBase API endpoint."""

    # @override
    # def load(self) -> dict[str, Any]:
    #     """Mock SessionBase API endpoint."""
    #     pass

    # @override
    # @classmethod
    # def clear_expired(cls) -> None:
    #     """Mock SessionBase API endpoint."""
    #     pass


class SignonTestCase(TestCase):
    """Base class for testing Signon and its subclasses."""

    signon_class: type[Signon] = Signon

    @override
    def setUp(self) -> None:
        """Provide a mock unauthenticated request for tests."""
        super().setUp()
        self.factory = RequestFactory()
        self.request = self.factory.get("/")
        self.request.session = MockSession()
        self.request.user = AnonymousUser()
        setattr(self.request, "_messages", FallbackStorage(self.request))
        self.providers: list[providers.Provider] = []
        self.enterContext(override_settings(SIGNON_PROVIDERS=self.providers))
        setattr(self.request, "signon", self.signon_class(self.request))

    def add_generic_provider(self, name: str = "generic") -> providers.Provider:
        """Add a generic provider to the provider configuration."""
        provider = providers.Provider(name=name, label=name.capitalize())
        self.providers.append(provider)
        return provider

    def make_gitlab_provider(
        self,
        name: str = "gitlab",
        restrict: Collection[str] | None = None,
        add_to_group: Any = None,
    ) -> providers.GitlabProvider:
        """Create a GitLab provider."""
        kwargs: dict[str, Any] = {}
        if restrict is not None:
            kwargs["restrict"] = restrict
        provider = providers.GitlabProvider(
            name=name,
            label=name.capitalize(),
            icon="signon/gitlabian.svg",
            client_id="123client_id",
            client_secret="123client_secret",
            url="https://salsa.debian.org",
            scope=("openid", "profile", "email"),
            **kwargs,
        )
        if add_to_group is not None:
            provider.options["add_to_group"] = add_to_group
        return provider

    def add_oidc_provider(
        self, restrict: Collection[str] | None = None
    ) -> providers.OIDCProvider:
        """Add a OIDC provider to the provider configuration."""
        kwargs: dict[str, Any] = {}
        if restrict is not None:
            kwargs["restrict"] = restrict
        provider = providers.OIDCProvider(
            name="oidc",
            label="OIDC",
            icon="signon/gitlabian.svg",
            client_id="123client_id",
            client_secret="123client_secret",
            url_issuer="https://example.com/",
            url_authorize="https://example.com/oauth/authorize",
            url_token="https://example.com/oauth/token",
            url_userinfo="https://example.com/oauth/userinfo",
            url_jwks="https://example.com/oauth/discovery/keys",
            scope=("openid", "profile", "email"),
            **kwargs,
        )
        self.providers.append(provider)
        return provider

    def add_gitlab_provider(
        self,
        name: str = "gitlab",
        restrict: Collection[str] | None = None,
        add_to_group: Any = None,
    ) -> providers.GitlabProvider:
        """Add a GitLab provider to the provider configuration."""
        provider = self.make_gitlab_provider(
            name=name, restrict=restrict, add_to_group=add_to_group
        )
        self.providers.append(provider)
        return provider

    def add_salsa_provider(
        self, restrict: Collection[str] | None = None, add_to_group: Any = None
    ) -> providers.GitlabProvider:
        """Add a Salsa provider to the provider configuration."""
        provider = self.make_gitlab_provider(
            name="salsa", restrict=restrict, add_to_group=add_to_group
        )
        self.providers.append(provider)
        return provider

    def add_entra_provider(
        self, add_to_group: Any = None
    ) -> providers.AzureEntraProvider:
        """Add an Azure Entra provider to the provider configuration."""
        provider = providers.AzureEntraProvider(
            name="entra",
            label="Azure Entra",
            client_id="123client_id",
            client_secret="123client_secret",
            tenant_id="tenant-id",
        )
        if add_to_group is not None:
            provider.options["add_to_group"] = add_to_group
        self.providers.append(provider)
        return provider

    def make_identity(
        self,
        user: User | None = None,
        issuer: str = "salsa",
        subject: str = "123",
        claims: dict[str, Any] | None = None,
    ) -> Identity:
        """Create a test identity."""
        if claims is None:
            claims = {}
        return Identity.objects.create(
            user=user,
            issuer=issuer,
            subject=subject,
            claims=claims,
        )


class SignonTests(SignonTestCase):
    """Test Signon."""

    def assertRestrictValidates(
        self, identity: Identity, restrict: Collection[str]
    ) -> None:
        """Identity passes this restrict rule."""
        assert isinstance(self.request, RequestSignonProtocol)
        provider = self.make_gitlab_provider(name=identity.issuer)
        provider.restrict = restrict
        self.assertEqual(
            self.request.signon.validate_claims(provider, identity), []
        )

    def assertRestrictDoesNotValidate(
        self,
        identity: Identity,
        restrict: Collection[str],
        *regexps: str,
    ) -> None:
        """Identity does not pass this restrict rule."""
        assert isinstance(self.request, RequestSignonProtocol)
        provider = self.make_gitlab_provider(name=identity.issuer)
        provider.restrict = restrict
        errors = self.request.signon.validate_claims(provider, identity)
        self.assertNotEqual(errors, [])
        for regexp in regexps:
            for idx, error in enumerate(errors):
                if re.search(regexp, error):  # pragma: no cover
                    errors.pop(idx)
                    break
            else:
                self.fail(
                    f"error list {errors!r} did not match {regexp!r}"
                )  # pragma: no cover
        if errors:
            self.fail(f"unexpected errors: {errors!r}")  # pragma: no cover

    def test_no_active_identities(self) -> None:
        """No active external identities leave request unauthenticated."""
        self.add_generic_provider()
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        self.assertEqual(self.request.signon.identities, {})
        self.assertFalse(self.request.user.is_authenticated)

        status = list(self.request.signon.status())
        self.assertEqual(len(status), 2)
        self.assertEqual(status[0][0].name, "generic")
        self.assertIsNone(status[0][1])
        self.assertEqual(status[1][0].name, "salsa")
        self.assertIsNone(status[1][1])

    def test_one_active_unbound_identity_autocreates(self) -> None:
        """One active but unbound identity tries to autocreate a user."""
        self.add_generic_provider()
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        identity = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@example.org",
                "email_verified": True,
            }
        )
        with self.assertLogs("debusine.server.signon") as log:
            self.assertTrue(self.request.signon.activate_identity(identity))

        self.assertEqual(
            self.request.signon.identities,
            {
                "salsa": identity,
            },
        )
        self.assertTrue(self.request.user.is_authenticated)

        identity.refresh_from_db()
        user = identity.user
        assert user is not None

        self.assertQuerySetEqual(user.identities.all(), [identity])
        self.assertEqual(user.username, "test@example.org")
        self.assertEqual(user.email, "test@example.org")
        self.assertEqual(user.first_name, "Test")
        self.assertEqual(user.last_name, "User")

        self.assertEqual(
            log.output,
            [
                "INFO:debusine.server.signon:"
                f"{user}: auto created from identity {identity}",
                "INFO:debusine.server.signon:"
                f"{user}: bound to identity {identity}",
            ],
        )

        status = list(self.request.signon.status())
        self.assertEqual(len(status), 2)
        self.assertEqual(status[0][0].name, "generic")
        self.assertIsNone(status[0][1])
        self.assertEqual(status[1][0].name, "salsa")
        self.assertEqual(status[1][1], identity)

    def test_one_active_bound_identity_logs_in(self) -> None:
        """One active bound identity is enough to authenticate."""
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user("test")
        identity = self.make_identity(user=user)
        with self.assertNoLogs("debusine.server.signon"):
            self.assertTrue(self.request.signon.activate_identity(identity))

        self.assertEqual(
            self.request.signon.identities,
            {
                "salsa": identity,
            },
        )
        self.assertEqual(self.request.user, user)
        backend_path = self.request.session.get(auth.BACKEND_SESSION_KEY)
        assert backend_path is not None
        self.assertIsInstance(
            auth.load_backend(backend_path), SignonAuthBackend
        )

    def test_add_aligned_bound_identity(self) -> None:
        """Adding a bound matching identity raises PermissionDenied."""
        self.add_generic_provider()
        self.add_salsa_provider()
        # Multiple active bound identities pointing to the same user
        # authenticate successfully
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user(
            "test", email="test@example.org"
        )
        ident1 = self.make_identity(user=user, issuer="generic")
        ident2 = self.make_identity(user=user, issuer="salsa")

        with self.assertNoLogs("debusine.server.signon"):
            self.assertTrue(self.request.signon.activate_identity(ident1))

        self.assertEqual(self.request.signon.identities, {"generic": ident1})
        self.assertEqual(self.request.user, user)

        with self.assertRaises(PermissionDenied):
            self.request.signon.activate_identity(ident2)

        self.assertEqual(
            self.request.signon.identities,
            {
                "generic": ident1,
            },
        )
        self.assertEqual(self.request.user, user)

    def test_add_unbound_identity(self) -> None:
        """Adding an unbound identity raises PermissionDenied."""
        self.add_generic_provider()
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user("test")
        ident1 = self.make_identity(user=user, issuer="generic")
        ident2 = self.make_identity(issuer="salsa")
        with self.assertNoLogs("debusine.server.signon"):
            self.assertTrue(self.request.signon.activate_identity(ident1))

        self.assertEqual(self.request.signon.identities, {"generic": ident1})
        self.assertEqual(self.request.user, user)

        with self.assertRaises(PermissionDenied):
            self.request.signon.activate_identity(ident2)

        self.assertEqual(self.request.signon.identities, {"generic": ident1})
        self.assertTrue(self.request.user.is_authenticated)
        self.assertEqual(self.request.user, user)

        ident1.refresh_from_db()
        self.assertEqual(ident1.user, user)
        ident2.refresh_from_db()
        self.assertIsNone(ident2.user)

    def test_add_conflicting_bound_identity(self) -> None:
        """Adding a bound conflicting identity raises PermissionDenied."""
        self.add_generic_provider()
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        user1 = get_user_model().objects.create_user(
            "test1", email="test1@example.org"
        )
        user2 = get_user_model().objects.create_user(
            "test2", email="test2@example.org"
        )
        ident1 = self.make_identity(user=user1, issuer="generic")
        ident2 = self.make_identity(user=user2, issuer="salsa")
        with self.assertNoLogs("debusine.server.signon"):
            self.assertTrue(self.request.signon.activate_identity(ident1))

        self.assertEqual(self.request.signon.identities, {"generic": ident1})
        self.assertEqual(self.request.user, user1)

        with self.assertRaises(PermissionDenied):
            self.request.signon.activate_identity(ident2)

        self.assertEqual(self.request.signon.identities, {"generic": ident1})
        self.assertTrue(self.request.user.is_authenticated)
        self.assertEqual(self.request.user, user1)

        ident1.refresh_from_db()
        self.assertEqual(ident1.user, user1)
        ident2.refresh_from_db()
        self.assertEqual(ident2.user, user2)

    def test_one_active_missing_identity(self) -> None:
        """One identity present in session but not the DB is removed."""
        self.add_generic_provider()
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        self.request.session["signon_identity_salsa"] = 1
        self.request.signon._compute_identities()

        self.assertEqual(self.request.signon.identities, {})
        self.assertFalse(self.request.user.is_authenticated)

        self.assertIsNone(self.request.session.get("signon_identity_salsa"))

        status = list(self.request.signon.status())
        self.assertEqual(len(status), 2)
        self.assertEqual(status[0][0].name, "generic")
        self.assertIsNone(status[0][1])
        self.assertEqual(status[1][0].name, "salsa")
        self.assertIsNone(status[1][1])

    def test_bind_intention_succeeds(self) -> None:
        """Bind intention binds to current user."""
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user("test")
        self.request.user = user

        ident = self.make_identity()
        with self.assertLogs("debusine.server.signon") as log:
            self.assertTrue(
                self.request.signon.activate_identity(ident, "bind")
            )

        self.assertEqual(self.request.signon.identities, {"salsa": ident})
        self.assertEqual(self.request.user, user)

        ident.refresh_from_db()
        self.assertEqual(ident.user, user)

        self.assertEqual(
            log.output,
            [f"INFO:debusine.server.signon:{user}: auto associated to {ident}"],
        )

    def test_bind_intention_succeeds_other_backend(self) -> None:
        """Bind intention binds to current user from a different backend."""
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user("test")

        # Simulate a valid existing login from a different auth backend
        self.request.user = user
        self.request.session[auth.BACKEND_SESSION_KEY] = (
            "django.contrib.auth.backends.ModelBackend"
        )

        ident = self.make_identity()
        with self.assertLogs("debusine.server.signon") as log:
            self.assertTrue(
                self.request.signon.activate_identity(ident, "bind")
            )

        self.assertEqual(self.request.signon.identities, {"salsa": ident})
        self.assertEqual(self.request.user, user)

        ident.refresh_from_db()
        self.assertEqual(ident.user, user)

        self.assertEqual(
            log.output,
            [
                "INFO:debusine.server.signon:"
                f"{user}: auto associated to {ident}",
            ],
        )

        ident.refresh_from_db()
        self.assertEqual(ident.user, user)

    def test_add_unbound_identities_binds(self) -> None:
        """One active bound identity and one unbound, when binding, binds."""
        self.add_generic_provider()
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user("test")
        ident1 = self.make_identity(user=user, issuer="generic")
        with self.assertNoLogs("debusine.server.signon"):
            self.assertTrue(self.request.signon.activate_identity(ident1))

        ident2 = self.make_identity(issuer="salsa")
        with self.assertLogs("debusine.server.signon") as log:
            self.assertTrue(
                self.request.signon.activate_identity(ident2, "bind")
            )

        self.assertEqual(
            self.request.signon.identities,
            {
                "generic": ident1,
                "salsa": ident2,
            },
        )
        self.assertEqual(self.request.user, user)

        ident1.refresh_from_db()
        self.assertEqual(ident1.user, user)
        ident2.refresh_from_db()
        self.assertEqual(ident2.user, user)

        self.assertEqual(
            log.output,
            [
                "INFO:debusine.server.signon:"
                f"{user}: auto associated to {ident2}",
            ],
        )

    def test_conflicting_bound_identities_bind(self) -> None:
        """Binding a second previously bound identity, rebinds."""
        self.add_generic_provider()
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        user1 = get_user_model().objects.create_user(
            "test1", email="test1@example.org"
        )
        ident1 = self.make_identity(user=user1, issuer="generic")
        with self.assertNoLogs("debusine.server.signon"):
            self.assertTrue(self.request.signon.activate_identity(ident1))

        user2 = get_user_model().objects.create_user(
            "test2", email="test2@example.org"
        )
        ident2 = self.make_identity(user=user2, issuer="salsa")
        with self.assertLogs("debusine.server.signon") as log:
            self.assertTrue(
                self.request.signon.activate_identity(ident2, "bind")
            )

        self.assertEqual(
            self.request.signon.identities,
            {
                "generic": ident1,
                "salsa": ident2,
            },
        )
        self.assertEqual(self.request.user, user1)

        ident1.refresh_from_db()
        self.assertEqual(ident1.user, user1)
        ident2.refresh_from_db()
        self.assertEqual(ident2.user, user1)

        self.assertEqual(
            log.output,
            [
                "INFO:debusine.server.signon:"
                f"{user1}: auto associated to {ident2}",
            ],
        )

    def test_logout_identities_logs_our_user_out(self) -> None:
        """logout_identities() deactivates identities."""
        # Log in using an external identity
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user("test")
        ident1 = self.make_identity(user=user)
        with self.assertNoLogs("debusine.server.signon"):
            self.assertTrue(self.request.signon.activate_identity(ident1))
        self.assertEqual(self.request.user, user)

        # Log out identities
        self.request.signon.logout_identities()
        self.assertEqual(self.request.signon.identities, {})

        # Since the user was logged in by external identities, it's now logged
        # out
        self.assertFalse(self.request.user.is_authenticated)

    def test_logout_identities_keep_user_from_other_backends(self) -> None:
        """logout_identities() deactivates identities."""
        # Log in using ModelBackend
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user("test")
        self.request.user = user
        self.request.session[auth.BACKEND_SESSION_KEY] = (
            "django.contrib.auth.backends.ModelBackend"
        )

        # Also activate an identity mapped to user
        ident1 = self.make_identity(user=user)
        with self.assertLogs("debusine.server.signon"):
            self.assertTrue(
                self.request.signon.activate_identity(ident1, "bind")
            )
        self.assertEqual(self.request.user, user)

        # Log out identities
        self.request.signon.logout_identities()
        self.assertEqual(self.request.signon.identities, {})

        # Since the user was logged in by another auth backend, it's still
        # logged in
        self.assertTrue(self.request.user.is_authenticated)
        self.assertEqual(self.request.user, user)

    def test_map_user_email(self) -> None:
        """Map identity to user based on email."""
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        user = get_user_model().objects.create_user(
            "test", email="test@example.org"
        )
        ident = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@example.org",
                "email_verified": True,
            }
        )

        with self.assertLogs("debusine.server.signon") as log:
            self.assertTrue(self.request.signon.activate_identity(ident))

        ident.refresh_from_db()
        self.assertTrue(self.request.user.is_authenticated)
        self.assertEqual(ident.user, user)

        self.assertEqual(
            log.output,
            [
                "INFO:debusine.server.signon:"
                f"{user}: user matched to identity {ident}",
                "INFO:debusine.server.signon:"
                f"{user}: bound to identity {ident}",
            ],
        )

    def test_map_user_unverified_email(self) -> None:
        """Do not map an identity with unverified email."""
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        get_user_model().objects.create_user("test", email="test@example.org")
        ident = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@example.org",
                "email_verified": False,
            }
        )

        with self.assertLogs("debusine.server.signon") as log:
            self.assertFalse(self.request.signon.activate_identity(ident))

        ident.refresh_from_db()
        self.assertFalse(self.request.user.is_authenticated)
        self.assertIsNone(ident.user)

        self.assertEqual(
            log.output,
            [
                "WARNING:debusine.server.signon:"
                f"identity {ident} does not have a verified email"
            ],
        )

    def test_map_user_invalid_email(self) -> None:
        """Test mapping an identity with an invalid email."""
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        ident = self.make_identity(
            claims={
                "name": "Test User",
                "email": "Invalid Email In Claim" * 20,
                "email_verified": True,
            }
        )

        with self.assertLogs("debusine.server.signon") as log:
            self.assertFalse(self.request.signon.activate_identity(ident))

        ident.refresh_from_db()
        self.assertFalse(self.request.user.is_authenticated)
        self.assertIsNone(ident.user)

        self.assertRegex(
            log.output[0],
            r"(?s)WARNING:debusine\.server\.signon:"
            f"{re.escape(str(ident))}: cannot create a local user"
            r".+Enter a valid username"
            r".+Enter a valid email",
        )
        self.assertEqual(len(log.output), 1)

    def test_auto_create_user_fails(self) -> None:
        """Do not autocreate users with unverified emails."""
        self.add_salsa_provider()
        assert isinstance(self.request, RequestSignonProtocol)
        ident = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@example.org",
                "email_verified": False,
            }
        )

        with self.assertLogs("debusine.server.signon") as log:
            self.assertFalse(self.request.signon.activate_identity(ident))

        ident.refresh_from_db()
        self.assertFalse(self.request.user.is_authenticated)
        self.assertIsNone(ident.user)

        self.assertEqual(
            log.output,
            [
                "WARNING:debusine.server.signon:"
                f"identity {ident} does not have a verified email"
            ],
        )

    def test_map_identity_invalid_provider(self) -> None:
        """An invalid provider in identity fails autocreation."""
        assert isinstance(self.request, RequestSignonProtocol)
        ident = self.make_identity(
            issuer="invalid",
            claims={
                "name": "Test User",
                "email": "test@example.org",
                "email_verified": False,
                "groups_direct": ["debian"],
            },
        )

        with (
            self.assertLogs("debusine.server.signon") as log,
            self.assertRaisesRegex(
                MapIdentityFailed, r"Invalid provider 'invalid' in identity"
            ),
        ):
            self.request.signon.identity_create_user(ident)

        self.assertEqual(
            log.output,
            [
                "WARNING:debusine.server.signon:"
                f"identity {ident} has unknown issuer invalid"
            ],
        )

    def test_default_email_verified(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        for label, create_provider, expected_verify_failure in (
            ("generic", self.add_generic_provider, False),
            ("oidc", self.add_oidc_provider, True),
            (
                "oidc-empty-restrict",
                lambda: self.add_oidc_provider(restrict=[]),
                False,
            ),
            ("entra", self.add_entra_provider, False),
        ):
            with self.subTest(label=label):
                try:
                    provider = create_provider()
                    identity = self.make_identity(
                        claims={
                            "name": "Test User",
                            "email": "test@example.org",
                            "email_verified": False,
                        },
                        issuer=provider.name,
                    )
                    expected = []
                    if expected_verify_failure:
                        expected.append(
                            f"identity {provider.name}:123 does not have a "
                            f"verified email"
                        )
                    self.assertEqual(
                        self.request.signon.validate_claims(provider, identity),
                        expected,
                    )
                finally:
                    identity.delete()

    def test_validate_empty_restrict(self) -> None:
        ident = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@example.org",
            }
        )
        self.assertRestrictValidates(ident, [])

    def test_validate_email_verified(self) -> None:
        """email-verified validates as intended."""
        ident = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@example.org",
                "email_verified": False,
            }
        )
        self.assertRestrictValidates(ident, [])
        self.assertRestrictDoesNotValidate(
            ident, ["email-verified"], "does not have a verified email"
        )

        ident.claims["email_verified"] = True
        self.assertRestrictValidates(ident, [])
        self.assertRestrictValidates(ident, ["email-verified"])

    def test_validate_group(self) -> None:
        """group:* validates as intended."""
        ident = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@example.org",
                "groups_direct": ["debian"],
            }
        )
        self.assertRestrictDoesNotValidate(
            ident, ["group:admin"], "not in group 'admin'"
        )
        self.assertRestrictValidates(ident, ["group:debian"])
        self.assertRestrictValidates(ident, ["group:debian,admin"])
        self.assertRestrictDoesNotValidate(
            ident,
            ["group:foo,bar"],
            "not in group 'foo' or 'bar'",
        )
        self.assertRestrictDoesNotValidate(
            ident,
            ["group:foo,bar,baz"],
            "not in group 'foo', 'bar' or 'baz'",
        )

    def test_validate(self) -> None:
        ident = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@example.org",
            }
        )
        for email_verified, groups, restrict, expected_issues in (
            (False, [], ["group:debian"], ["not in group 'debian'"]),
            (False, ["debian"], ["group:debian"], []),
            (False, ["debian"], ["group:admin"], ["not in group 'admin'"]),
            (
                False,
                ["debian"],
                ["group:debian", "group:admin"],
                ["not in group 'admin'"],
            ),
            (
                False,
                [],
                ["group:debian", "group:admin"],
                ["not in group 'debian'", "not in group 'admin'"],
            ),
            (False, ["debian", "admin"], ["group:debian"], []),
            (False, ["debian", "admin"], ["group:admin"], []),
            (False, ["debian", "admin"], ["group:debian", "group:admin"], []),
            (
                False,
                ["debian"],
                ["email-verified"],
                ["does not have a verified email"],
            ),
            (
                False,
                ["debian"],
                ["email-verified", "group:debian"],
                ["does not have a verified email"],
            ),
            (True, ["debian"], ["email-verified"], []),
            (True, ["debian"], ["group:debian"], []),
            (True, ["debian"], ["email-verified", "group:debian"], []),
            (
                True,
                ["debian"],
                ["email-verified", "group:admin"],
                ["not in group 'admin'"],
            ),
            (
                True,
                ["debian", "admin"],
                ["email-verified", "group:debian,ubuntu", "group:admin"],
                [],
            ),
            (
                True,
                ["ubuntu", "admin"],
                ["email-verified", "group:debian,ubuntu", "group:admin"],
                [],
            ),
            (
                True,
                ["debian", "ubuntu"],
                ["email-verified", "group:debian,ubuntu", "group:admin"],
                ["not in group 'admin'"],
            ),
        ):
            with self.subTest(
                email_verified=True, groups=groups, restrict=restrict
            ):
                ident.claims["email_verified"] = email_verified
                ident.claims["groups_direct"] = groups
                if not expected_issues:
                    self.assertRestrictValidates(ident, restrict)
                else:
                    self.assertRestrictDoesNotValidate(
                        ident, restrict, *expected_issues
                    )

    def test_validate_invalid_expression(self) -> None:
        """An invalid restrict value raises ImproperlyConfigured."""
        assert isinstance(self.request, RequestSignonProtocol)
        provider = self.make_gitlab_provider(restrict=["invalid"])
        identity = self.make_identity()
        with self.assertRaises(ImproperlyConfigured):
            self.request.signon.validate_claims(provider, identity)

    def test_validate_claims_gitlab_not_contributor(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        provider = self.add_gitlab_provider(
            restrict=["group:freexian/teams/lts-contributors"]
        )
        identity = self.make_identity(issuer="gitlab")
        self.assertEqual(
            self.request.signon.validate_claims(provider, identity),
            [
                "identity gitlab:123 is not in group"
                " 'freexian/teams/lts-contributors'"
            ],
        )

    def test_validate_claims_gitlab_contributor(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        provider = self.add_gitlab_provider(
            restrict=["group:freexian/teams/lts-contributors"]
        )
        identity = self.make_identity(issuer="gitlab")
        identity.claims["groups_direct"] = ["freexian/teams/lts-contributors"]
        identity.save()
        self.assertEqual(
            self.request.signon.validate_claims(provider, identity),
            [],
        )

    def test_identity_create_user_unknown_provider(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        identity = self.make_identity(issuer="unknown")
        with self.assertRaisesRegex(
            MapIdentityFailed, r"Invalid provider 'unknown' in identity"
        ):
            self.request.signon.identity_create_user(identity)

    def test_create_user_unsupported_provider(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        identity = self.make_identity(
            issuer="nonexisting",
            claims={
                "name": "Test User",
                "email": "test@debian.org",
                "email_verified": True,
            },
        )
        user = self.request.signon.create_user_from_identity(identity)
        assert user is not None
        self.assertQuerySetEqual(user.debusine_groups.all(), [])

    def test_setup_new_user_missing_group(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        self.add_salsa_provider(add_to_group={"debian": "debusine/missing"})
        identity = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@debian.org",
                "email_verified": True,
                "groups_direct": ["debian"],
            }
        )
        user = self.request.signon.create_user_from_identity(identity)
        assert user is not None
        with self.assertRaisesRegex(
            ImproperlyConfigured, "Group 'debusine/missing' not found"
        ):
            self.request.signon.setup_new_user(user, identity)

    def test_setup_new_user_missing_provider(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        identity = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@debian.org",
                "email_verified": True,
                "groups_direct": ["debian"],
            }
        )
        user = self.request.signon.create_user_from_identity(identity)
        assert user is not None
        with mock.patch(
            "debusine.server.signon.signon.Signon.add_user_to_groups"
        ) as add_user_to_groups:
            self.request.signon.setup_new_user(user, identity)
        add_user_to_groups.assert_not_called()

    def test_setup_new_user_generic_provider(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        self.add_generic_provider()
        identity = self.make_identity(
            claims={
                "name": "Test User",
                "email": "test@debian.org",
            },
            issuer="generic",
        )
        user = self.request.signon.create_user_from_identity(identity)
        assert user is not None
        self.request.signon.setup_new_user(user, identity)
        self.assertFalse(user.debusine_groups.exists())

    def test_setup_new_user_add_to_group(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        debian_scope = self.playground.get_or_create_scope("debian")
        freexian_scope = self.playground.get_or_create_scope("freexian")
        debian_group = self.playground.create_group(
            name="Debian", scope=debian_scope
        )
        elts_group = self.playground.create_group(
            name="elts-contributors", scope=freexian_scope
        )
        groups = [debian_group, elts_group]

        user = get_user_model().objects.create_user("test")
        identity = self.make_identity(
            issuer="salsa",
            claims={
                "name": "Test User",
                "email": "test@debian.org",
                "email_verified": True,
                "groups_direct": ["debian"],
            },
        )
        self.add_salsa_provider()
        add_to_group = {
            "debian": "debian/Debian",
            "freexian/teams/lts-contributors": "freexian/elts-contributors",
            "nm:dd": "debian/Debian",
            "nm:dm": "debian/DM",
            "nm:dm_ga": "debian/DM",
        }
        signon = self.request.signon
        salsa_groups: None | list[str]
        for salsa_groups, debusine_groups in (
            (None, []),
            ([], []),
            (["debian"], [debian_group]),
            (["freexian/teams/lts-contributors"], [elts_group]),
            (
                ["debian", "freexian/teams/lts-contributors"],
                [debian_group, elts_group],
            ),
        ):
            with self.subTest(salsa_groups=salsa_groups):
                if salsa_groups is None:
                    identity.claims.pop("groups_direct", None)
                else:
                    identity.claims["groups_direct"] = salsa_groups
                identity.save()
                for group in groups:
                    group.users.remove(user)
                signon.add_user_to_groups(user, identity, add_to_group)
                try:
                    self.assertQuerySetEqual(
                        user.debusine_groups.all(),
                        debusine_groups,
                        ordered=False,
                    )
                finally:
                    identity.user = None
                    identity.save()

    def test_entra_add_to_group(self) -> None:
        assert isinstance(self.request, RequestSignonProtocol)
        admin_group = self.playground.create_group(name="Admins")
        user = get_user_model().objects.create_user("test")
        admin_guid = "45ab6a0d-cf51-49c6-b340-661854e56fed"
        identity = self.make_identity(
            issuer="entra",
            claims={
                "name": "Test User",
                "email": "test@debian.org",
                "groups": [admin_guid],
            },
        )
        self.add_entra_provider(add_to_group={admin_guid: str(admin_group)})
        signon = self.request.signon
        signon.setup_new_user(user, identity)
        self.assertQuerySetEqual(
            user.debusine_groups.all(),
            [admin_group],
            ordered=False,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the blhc workflow."""

import datetime as dt
from collections.abc import Sequence
from typing import Any, ClassVar, NewType, override
from unittest import mock

from django.test import override_settings
from django.utils import timezone

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianBlhc,
    DebianSourcePackage,
    TaskTypes,
)
from debusine.client.models import LookupChildType, LookupResultType
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    WorkRequest,
    Workspace,
)
from debusine.db.models.lookups import LookupResult
from debusine.db.models.work_requests import SkipWorkRequest
from debusine.server.workflows import BlhcWorkflow, WorkflowValidationError
from debusine.server.workflows.models import (
    BaseWorkflowData,
    BlhcWorkflowData,
    BlhcWorkflowDynamicData,
    RegressionTrackingUpdateQAResults,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.server.workflows.tests.test_regression_tracking import (
    RegressionTrackingWorkflowTestMixin,
)
from debusine.tasks.models import (
    ActionSkipIfLookupResultChanged,
    ActionUpdateCollectionWithArtifacts,
    BaseDynamicTaskData,
    BlhcFlags,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    OutputData,
    RegressionAnalysisStatus,
    SbuildData,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry

Architecture = NewType("Architecture", str)


class BlhcWorkflowTests(
    RegressionTrackingWorkflowTestMixin, WorkflowTestCase[BlhcWorkflow]
):
    """Unit tests for :py:class:`BlhcWorkflow`."""

    package_build_log_artifacts: ClassVar[dict[Architecture, Artifact]]
    source_artifact: ClassVar[Artifact]

    workspace: ClassVar[Workspace]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.workspace = cls.playground.get_default_workspace()

        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello", version="1.0-1"
        )

        cls.package_build_log_artifacts = {}
        for arch in ["amd64", "arm64", "all"]:
            artifact = cls.playground.create_build_log_artifact(
                source="hello", version="1.0-1", build_arch=arch
            )

            cls.package_build_log_artifacts[Architecture(arch)] = artifact

    def orchestrate(
        self,
        *,
        task_data: BlhcWorkflowData,
        source_artifact: Artifact,
        architectures: Sequence[str] | None = None,
        parent: WorkRequest | None = None,
        pipeline_task_name: str = "examplepipeline",
    ) -> WorkRequest:
        """Create and orchestrate a BlhcWorkflow."""
        if architectures is None:
            architectures = ["amd64", "arm64", "all"]

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Pipeline workflow."""

            TASK_NAME = pipeline_task_name

            @override
            def populate(self_) -> None:
                source_data = source_artifact.create_data()
                assert isinstance(source_data, DebianSourcePackage)
                assert architectures is not None

                sbuild, _ = self_.work_request_ensure_child_workflow(
                    task_name="sbuild",
                    task_data=SbuildWorkflowData(
                        input=SbuildInput(source_artifact=source_artifact.id),
                        target_distribution="debian:trixie",
                        architectures=list(architectures),
                    ),
                    workflow_data=WorkRequestWorkflowData(
                        display_name="sbuild", step="sbuild"
                    ),
                )
                self.playground.advance_work_request(sbuild, mark_running=True)
                for arch in architectures:
                    child = sbuild.create_child_worker(
                        task_name="sbuild",
                        task_data=SbuildData(
                            input=SbuildInput(
                                source_artifact=source_artifact.id
                            ),
                            build_architecture=arch,
                            environment="debian/match:codename=trixie",
                        ),
                    )
                    self_.provides_artifact(
                        child,
                        ArtifactCategory.PACKAGE_BUILD_LOG,
                        name=f"buildlog-{arch}",
                        data={
                            "source_package_name": source_data.name,
                            "source_package_version": source_data.version,
                            "architecture": arch,
                        },
                    )
                sbuild.unblock_workflow_children()

                blhc, _ = self_.work_request_ensure_child_workflow(
                    task_name="blhc",
                    task_data=task_data,
                    workflow_data=WorkRequestWorkflowData(
                        display_name="blhc",
                        step="blhc",
                    ),
                )
                self_.orchestrate_child(blhc)

        root = self.playground.create_workflow(
            task_name=pipeline_task_name, parent=parent
        )
        self.assertTrue(
            self.playground.orchestrate_workflow(root, use_data=False)
        )

        return root

    def setup_blhc_workflow(
        self,
        *,
        binary_artifacts: LookupMultiple | None = None,
        package_build_logs: LookupMultiple | None = None,
        extra_task_data: dict[str, Any] | None = None,
        include_source_artifact: bool = True,
    ) -> None:
        """
        Create and schedule a Blhc workflow.

        :param binary_artifacts: binary artifacts for the workflow.
        :param package_build_logs: build logs for the workflow. If both
          ``binary_artifacts`` and ``package_build_logs`` are None, use all
          the build log artifacts created by ``setUpTestData()``.
        :param extra_task_data: any extra task data to pass to the workflow.
        :param include_source_artifact: True to include self.source_artifact.
        """
        if binary_artifacts is None and package_build_logs is None:
            package_build_logs = LookupMultiple(
                tuple(
                    package_build_log.id
                    for package_build_log in (
                        self.package_build_log_artifacts.values()
                    )
                )
            )

        task_data = BlhcWorkflowData(
            source_artifact=(
                self.source_artifact.id if include_source_artifact else None
            ),
            binary_artifacts=binary_artifacts,
            package_build_logs=package_build_logs,
            vendor="debian",
            codename="bookworm",
        )

        if extra_task_data is not None:
            task_data = task_data.model_copy(update=extra_task_data)

        example_pipeline = self.orchestrate(
            task_data=task_data,
            source_artifact=self.source_artifact,
        )
        work_request = example_pipeline.children.get(
            task_name="blhc", task_type=TaskTypes.WORKFLOW
        )
        self.task_set_current(self.get_workflow(work_request))

    def add_qa_result(
        self,
        qa_results: Collection,
        package: str,
        version: str,
        architecture: str,
        *,
        timestamp: dt.datetime | None = None,
    ) -> CollectionItem:
        """Add a blhc result to a ``debian:qa-results`` collection."""
        work_request = self.playground.create_worker_task(
            task_name="blhc",
            result=WorkRequest.Results.SUCCESS,
            validate=False,
            use_data=False,
        )
        return qa_results.manager.add_artifact(
            self.playground.create_artifact(
                category=ArtifactCategory.BLHC,
                data=DebianBlhc(
                    source=package, version=version, architecture=architecture
                ),
            )[0],
            user=self.playground.get_default_user(),
            variables={
                "package": package,
                "version": version,
                "architecture": architecture,
                "timestamp": int((timestamp or timezone.now()).timestamp()),
                "work_request_id": work_request.id,
            },
        )

    @preserve_db_task_registry()
    def test_validate_input(self) -> None:
        """validate_input passes a valid case."""
        # validate_input() is called and does not raise
        self.setup_blhc_workflow()

    @preserve_db_task_registry()
    def test_validate_input_bad_qa_suite(self) -> None:
        """validate_input raises errors in looking up a suite."""
        self.setup_blhc_workflow(
            extra_task_data={"qa_suite": "nonexistent@debian:suite"}
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:suite' does not exist or is hidden",
        ):
            self.task.validate_input()

    @preserve_db_task_registry()
    def test_validate_input_bad_reference_qa_results(self) -> None:
        """validate_input raises errors in looking up reference QA results."""
        self.setup_blhc_workflow(
            extra_task_data={
                "reference_qa_results": "nonexistent@debian:qa-results"
            }
        )

        with self.assertRaisesRegex(
            WorkflowValidationError,
            "'nonexistent@debian:qa-results' does not exist or is hidden",
        ):
            self.task.validate_input()

    def test_has_current_reference_qa_result_no_match(self) -> None:
        """_has_current_reference_qa_result: no matching result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "other", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                vendor="debian",
                codename="sid",
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.ARTIFACT,
                    lookup=build_log.id,
                    artifact=build_log,
                ),
                "amd64",
            )
        )

    def test_has_current_reference_qa_result_different_version(self) -> None:
        """_has_current_reference_qa_result: result for different version."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-2"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                vendor="debian",
                codename="sid",
            ),
        )

        self.assertFalse(
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.ARTIFACT,
                    lookup=build_log.id,
                    artifact=build_log,
                ),
                "amd64",
            )
        )

    def test_has_current_reference_qa_result_unexpected(self) -> None:
        """_has_current_reference_qa_result: unexpected result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                vendor="debian",
                codename="sid",
            ),
        )

        with self.assertRaisesRegex(ValueError, r"^Unexpected result"):
            artifact, _ = self.playground.create_artifact()
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.ARTIFACT,
                    lookup=artifact.id,
                    artifact=artifact,
                ),
                "amd64",
            )

    def test_has_current_reference_qa_result_current(self) -> None:
        """_has_current_reference_qa_result: current result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                vendor="debian",
                codename="sid",
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.ARTIFACT,
                    lookup=build_log.id,
                    artifact=build_log,
                ),
                "amd64",
            )
        )

    def test_has_current_reference_qa_result_current_promise(self) -> None:
        """_has_current_reference_qa_result: current result matching promise."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )
        build_log_promise = self.playground.create_bare_data_item(
            collection,
            "test",
            category=BareDataCategory.PROMISE,
            data={
                "promise_work_request_id": 2,
                "promise_workflow_id": 1,
                "promise_category": ArtifactCategory.PACKAGE_BUILD_LOG,
                "source_package_name": "hello",
                "source_package_version": "1.0-1",
            },
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple.model_validate(
                    [f"test@{CollectionCategory.TEST}/name:test"]
                ),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                vendor="debian",
                codename="sid",
            ),
        )

        self.assertTrue(
            self.get_workflow(wr)._has_current_reference_qa_result(
                LookupResult(
                    result_type=LookupResultType.BARE,
                    lookup=f"test@{CollectionCategory.TEST}/test",
                    collection_item=build_log_promise,
                ),
                "amd64",
            )
        )

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        self.setup_blhc_workflow(
            extra_task_data={
                "extra_flags": [BlhcFlags.BINDNOW, BlhcFlags.LINE_NUMBERS]
            }
        )

        # One work request per architecture
        self.assertQuerySetEqual(
            self.work_request.children.order_by("id").values_list(
                "task_type", "task_name", "task_data", "workflow_data_json"
            ),
            [
                (
                    TaskTypes.WORKER,
                    "blhc",
                    {
                        "environment": "debian/match:codename=bookworm",
                        "build_architecture": (
                            "amd64" if architecture == "all" else architecture
                        ),
                        "input": {"artifact": artifact.id},
                        "extra_flags": [
                            BlhcFlags.BINDNOW,
                            BlhcFlags.LINE_NUMBERS,
                        ],
                    },
                    {
                        "display_name": (
                            f"build log hardening check for {architecture}"
                        ),
                        "step": f"blhc-{architecture}",
                    },
                )
                for architecture, artifact in sorted(
                    self.package_build_log_artifacts.items()
                )
            ],
        )

    def test_populate_with_uploads(self) -> None:
        source_upload = self.playground.create_upload_artifacts(
            binary=False
        ).upload
        binary_upload = self.playground.create_upload_artifacts(
            source=False, binaries=[("hello", "amd64")]
        ).upload
        log = self.package_build_log_artifacts[Architecture("amd64")]
        self.playground.create_artifact_relation(
            artifact=log, target=binary_upload
        )
        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                source_artifact=source_upload.id,
                binary_artifacts=LookupMultiple((binary_upload.id,)),
                vendor="debian",
                codename="sid",
            ),
        )
        self.get_workflow(wr).populate()
        children = list(wr.children.all())
        self.assertEqual(len(children), 1)
        child = children[0]
        self.assertEqual(
            child.task_data["input"], {"artifact": f"{log.id}@artifacts"}
        )

    def test_populate_has_current_reference_qa_result(self) -> None:
        """The workflow does nothing with a current reference QA result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                vendor="debian",
                codename="sid",
            ),
        )

        self.get_workflow(wr).populate()

        self.assertQuerySetEqual(wr.children.all(), [])

    def test_populate_has_current_reference_qa_result_force(self) -> None:
        """The workflow can be forced despite a current reference QA result."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.add_qa_result(sid_qa_results, "hello", "1.0-1", "amd64")
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.FORCE,
                vendor="debian",
                codename="sid",
            ),
        )

        self.get_workflow(wr).populate()

        self.assertTrue(
            wr.children.filter(
                task_type=TaskTypes.WORKER, task_name="blhc"
            ).exists()
        )

    def test_populate_previous_reference_qa_result_outdated(self) -> None:
        """The workflow produces a reference QA result if it is outdated."""
        sid_suite = self.playground.create_collection(
            "sid", CollectionCategory.SUITE
        )
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        old_qa_result = self.add_qa_result(
            sid_qa_results, "hello", "1.0-1", "amd64"
        )
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-2"
        )
        source_item = sid_suite.manager.add_artifact(
            self.playground.create_source_artifact(
                name="hello", version="1.0-2"
            ),
            user=self.playground.get_default_user(),
            variables={"component": "main", "section": "devel"},
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                vendor="debian",
                codename="sid",
            ),
        )

        self.get_workflow(wr).populate()

        [child] = wr.children.all()
        self.assertEqual(child.status, WorkRequest.Statuses.BLOCKED)
        self.assertEqual(child.task_type, TaskTypes.WORKER)
        self.assertEqual(child.task_name, "blhc")
        self.assertEqual(
            child.task_data,
            {
                "environment": "debian/match:codename=sid",
                "build_architecture": "amd64",
                "input": {"artifact": build_log.id},
                "extra_flags": [],
            },
        )
        provide_action = ActionUpdateCollectionWithArtifacts(
            collection="internal@collections",
            name_template=("reference-qa-result|blhc-amd64"),
            artifact_filters={"category": ArtifactCategory.BLHC},
        )
        skip_action = ActionSkipIfLookupResultChanged(
            lookup=(
                f"sid@{CollectionCategory.QA_RESULTS}/latest:blhc_hello_amd64"
            ),
            collection_item_id=old_qa_result.id,
            promise_name="reference-qa-result|blhc-amd64",
        )
        qa_result_action = ActionUpdateCollectionWithArtifacts(
            collection=f"sid@{CollectionCategory.QA_RESULTS}",
            variables={
                "package": "hello",
                "version": "1.0-2",
                "architecture": "amd64",
                "timestamp": int(source_item.created_at.timestamp()),
                "work_request_id": child.id,
            },
            artifact_filters={"category": ArtifactCategory.BLHC},
        )
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[skip_action],
            on_failure=[qa_result_action],
            on_success=[provide_action, qa_result_action],
        )

        # Completing the work request stores the QA result.
        result, _ = self.playground.create_artifact(
            category=ArtifactCategory.BLHC,
            data=DebianBlhc(
                source="hello", version="1.0-2", architecture="amd64"
            ),
            work_request=child,
        )
        self.playground.advance_work_request(
            child, result=WorkRequest.Results.SUCCESS, use_data=False
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/latest:blhc_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).artifact,
            result,
        )

        # Population is idempotent, even if qa_suite changes.
        with (
            mock.patch.object(
                BlhcWorkflow,
                "_has_current_reference_qa_result",
                return_value=False,
            ),
            mock.patch.object(
                BlhcWorkflow,
                "qa_suite_changed",
                mock.PropertyMock(
                    return_value=timezone.now() + dt.timedelta(hours=1)
                ),
            ),
        ):
            self.get_workflow(wr).populate()
        self.assertEqual(wr.children.count(), 1)
        child.refresh_from_db()
        self.assert_work_request_event_reactions(
            child,
            on_assignment=[skip_action],
            on_failure=[qa_result_action],
            on_success=[provide_action, qa_result_action],
        )

    def test_populate_reference_qa_result_backs_off(self) -> None:
        """Reference tasks are skipped if another workflow got there first."""
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        sid_qa_results = self.playground.create_collection(
            "sid", CollectionCategory.QA_RESULTS
        )
        self.playground.create_debian_environment(
            codename="sid", variant="blhc"
        )
        build_log = self.playground.create_build_log_artifact(
            source="hello", version="1.0-1"
        )

        wr = self.playground.create_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                package_build_logs=LookupMultiple((build_log.id,)),
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                vendor="debian",
                codename="sid",
            ),
        )

        self.get_workflow(wr).populate()

        racing_qa_result = self.add_qa_result(
            sid_qa_results, "hello", "1.0-1", "amd64"
        )
        [child] = wr.children.all()
        self.playground.advance_work_request(child, mark_pending=True)

        with self.assertRaises(SkipWorkRequest):
            child.assign_worker(self.playground.create_worker())

        self.assertEqual(child.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(child.result, WorkRequest.Results.SKIPPED)
        self.assertEqual(
            child.output_data,
            OutputData(
                skip_reason=(
                    f"Result of lookup "
                    f"'sid@{CollectionCategory.QA_RESULTS}/"
                    f"latest:blhc_hello_amd64' changed"
                )
            ),
        )
        self.assertEqual(
            child.lookup_single(
                f"sid@{CollectionCategory.QA_RESULTS}/latest:blhc_hello_amd64",
                expect_type=LookupChildType.ARTIFACT,
            ).collection_item,
            racing_qa_result,
        )

    @override_settings(DISABLE_AUTOMATIC_SCHEDULING=True)
    @preserve_db_task_registry()
    def test_callback_regression_analysis(self) -> None:
        self.playground.create_collection("sid", CollectionCategory.SUITE)
        self.playground.create_collection("sid", CollectionCategory.QA_RESULTS)
        architectures = ["amd64", "arm64", "i386", "ppc64el", "s390x"]
        reference_source_artifact = self.playground.create_source_artifact(
            name="hello", version="0.9-1"
        )

        root = self.playground.create_workflow(mark_running=True)
        reference = self.orchestrate(
            task_data=BlhcWorkflowData(
                prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                update_qa_results=RegressionTrackingUpdateQAResults.YES,
                source_artifact=reference_source_artifact.id,
                package_build_logs=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:buildlog-{arch}"
                        for arch in architectures
                        if arch != "ppc64el"
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            source_artifact=reference_source_artifact,
            architectures=architectures,
            parent=root,
            pipeline_task_name="referencepipeline",
        )
        new = self.orchestrate(
            task_data=BlhcWorkflowData(
                reference_prefix="reference-qa-result|",
                qa_suite=f"sid@{CollectionCategory.SUITE}",
                reference_qa_results=f"sid@{CollectionCategory.QA_RESULTS}",
                enable_regression_tracking=True,
                source_artifact=self.source_artifact.id,
                package_build_logs=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:buildlog-{arch}"
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="sid",
            ),
            source_artifact=self.source_artifact,
            architectures=architectures,
            parent=root,
        )

        # Unblock the blhc tasks.
        for sbuild in reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )
        for sbuild in new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        ).children.filter(task_type=TaskTypes.WORKER, task_name="sbuild"):
            self.playground.advance_work_request(
                sbuild, result=WorkRequest.Results.SUCCESS, use_data=False
            )

        reference_blhc_workflow = reference.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="blhc"
        )
        reference_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in reference_blhc_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="blhc"
            )
        }
        reference_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.BLHC,
                data=DebianBlhc(
                    source="hello", version="0.9-1", architecture=architecture
                ),
                work_request=reference_tasks[architecture],
            )[0]
            for architecture in architectures
            if architecture != "ppc64el"
        }
        new_blhc_workflow = new.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="blhc"
        )
        new_tasks = {
            wr.task_data["build_architecture"]: wr
            for wr in new_blhc_workflow.children.filter(
                task_type=TaskTypes.WORKER, task_name="blhc"
            )
        }
        new_results = {
            architecture: self.playground.create_artifact(
                category=ArtifactCategory.BLHC,
                data=DebianBlhc(
                    source="hello", version="1.0-1", architecture=architecture
                ),
                work_request=new_tasks[architecture],
            )[0]
            for architecture in architectures
        }

        self.assertEqual(
            new_blhc_workflow.output_data, OutputData(regression_analysis={})
        )
        self.schedule_and_run_workflow_callback(new_blhc_workflow)
        self.playground.advance_work_request(
            reference_tasks["amd64"],
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )
        self.schedule_and_run_workflow_callback(new_blhc_workflow)
        self.playground.advance_work_request(
            new_tasks["amd64"],
            result=WorkRequest.Results.SUCCESS,
            use_data=False,
        )

        self.schedule_and_run_workflow_callback(new_blhc_workflow)

        expected_analysis: dict[str, tuple[RegressionAnalysisStatus, None]] = {
            "amd64": (RegressionAnalysisStatus.STABLE, None)
        }
        new_blhc_workflow.refresh_from_db()
        self.assert_regression_analysis(
            new_blhc_workflow,
            "0.9-1",
            reference_results,
            "1.0-1",
            new_results,
            expected_analysis,
        )

        self.assert_callback_status(
            new_blhc_workflow, WorkRequest.Statuses.BLOCKED
        )

        for architecture, reference_result, new_result, status in (
            (
                "arm64",
                WorkRequest.Results.SUCCESS,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.REGRESSION,
            ),
            (
                "i386",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.FAILURE,
                RegressionAnalysisStatus.STABLE,
            ),
            (
                "ppc64el",
                None,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.NO_RESULT,
            ),
            (
                "s390x",
                WorkRequest.Results.FAILURE,
                WorkRequest.Results.SUCCESS,
                RegressionAnalysisStatus.IMPROVEMENT,
            ),
        ):
            if reference_result is not None:
                self.playground.advance_work_request(
                    reference_tasks[architecture],
                    result=reference_result,
                    use_data=False,
                )
            self.playground.advance_work_request(
                new_tasks[architecture], result=new_result, use_data=False
            )
            expected_analysis[architecture] = (status, None)

        self.assert_callback_status(
            new_blhc_workflow, WorkRequest.Statuses.PENDING
        )

        self.schedule_and_run_workflow_callback(new_blhc_workflow)

        new_blhc_workflow.refresh_from_db()
        assert new_blhc_workflow.output_data is not None
        self.assert_regression_analysis(
            new_blhc_workflow,
            "0.9-1",
            reference_results,
            "1.0-1",
            new_results,
            expected_analysis,
        )
        self.assert_callback_status(
            new_blhc_workflow, WorkRequest.Statuses.COMPLETED
        )

    @preserve_db_task_registry()
    def test_compute_scheduler_tags(self) -> None:
        package_build_log_1 = self.playground.create_build_log_artifact(
            source="hello"
        )
        package_build_log_2 = self.playground.create_build_log_artifact(
            source="hello-traditional"
        )
        self.setup_blhc_workflow(
            package_build_logs=LookupMultiple(
                (
                    f"{package_build_log_1.id}@artifacts",
                    f"{package_build_log_2.id}@artifacts",
                )
            )
        )
        self.assertComputesSchedulerTags()

    @preserve_db_task_registry()
    def test_compute_dynamic_data_with_binary_artifacts(self) -> None:
        binary_uploads = {
            architecture: self.playground.create_upload_artifacts(
                source=False, binaries=[("hello", architecture)]
            ).upload
            for architecture in ("amd64", "s390x")
        }
        package_build_logs = {
            architecture: self.playground.create_build_log_artifact(
                build_arch=architecture
            )
            for architecture in binary_uploads
        }
        for architecture in binary_uploads:
            self.playground.create_artifact_relation(
                artifact=package_build_logs[architecture],
                target=binary_uploads[architecture],
            )

        self.setup_blhc_workflow(
            binary_artifacts=LookupMultiple(
                tuple(
                    f"{artifact.id}@artifacts"
                    for artifact in binary_uploads.values()
                )
            )
        )

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BlhcWorkflowDynamicData(
                subject="hello",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[
                    artifact.id for artifact in binary_uploads.values()
                ],
                package_build_logs_ids=sorted(
                    [artifact.id for artifact in package_build_logs.values()]
                ),
            ),
        )

    @preserve_db_task_registry()
    def test_compute_dynamic_data_with_package_build_logs(self) -> None:
        package_build_log_1 = self.playground.create_build_log_artifact(
            source="hello"
        )
        package_build_log_2 = self.playground.create_build_log_artifact(
            source="hello-traditional"
        )

        self.setup_blhc_workflow(
            package_build_logs=LookupMultiple(
                (
                    f"{package_build_log_1.id}@artifacts",
                    f"{package_build_log_2.id}@artifacts",
                )
            )
        )

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            BlhcWorkflowDynamicData(
                subject="hello hello-traditional",
                source_artifact_id=self.source_artifact.id,
                package_build_logs_ids=[
                    package_build_log_1.id,
                    package_build_log_2.id,
                ],
            ),
        )

    @preserve_db_task_registry()
    def test_get_input_artifacts_with_binary_artifacts(self) -> None:
        binary_uploads = {
            architecture: self.playground.create_upload_artifacts(
                source=False, binaries=[("hello", architecture)]
            ).upload
            for architecture in ("amd64", "s390x")
        }

        self.setup_blhc_workflow(
            binary_artifacts=LookupMultiple(
                tuple(
                    f"{artifact.id}@artifacts"
                    for artifact in binary_uploads.values()
                )
            )
        )

        assert self.task.data.source_artifact is not None
        assert self.task.data.binary_artifacts is not None
        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=self.task.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[
                        artifact.id for artifact in binary_uploads.values()
                    ],
                ),
            ],
        )

    @preserve_db_task_registry()
    def test_get_input_artifacts_with_package_build_logs(self) -> None:
        package_build_log_1 = self.playground.create_build_log_artifact(
            source="hello"
        )
        package_build_log_2 = self.playground.create_build_log_artifact(
            source="hello-traditional"
        )

        self.setup_blhc_workflow(
            package_build_logs=LookupMultiple(
                (
                    f"{package_build_log_1.id}@artifacts",
                    f"{package_build_log_2.id}@artifacts",
                )
            )
        )

        assert self.task.data.source_artifact is not None
        assert self.task.data.package_build_logs is not None
        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=self.task.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.package_build_logs,
                    label="package_build_logs",
                    artifact_ids=[
                        package_build_log_1.id,
                        package_build_log_2.id,
                    ],
                ),
            ],
        )

    @preserve_db_task_registry()
    def test_get_input_artifacts_without_source_artifact(self) -> None:
        package_build_log_1 = self.playground.create_build_log_artifact(
            source="hello"
        )
        package_build_log_2 = self.playground.create_build_log_artifact(
            source="hello-traditional"
        )

        self.setup_blhc_workflow(
            package_build_logs=LookupMultiple(
                (
                    f"{package_build_log_1.id}@artifacts",
                    f"{package_build_log_2.id}@artifacts",
                )
            ),
            include_source_artifact=False,
        )

        assert self.task.data.package_build_logs is not None
        self.assertEqual(
            self.task.get_input_artifacts(),
            [
                InputArtifactMultiple.from_ids(
                    lookup=self.task.data.package_build_logs,
                    label="package_build_logs",
                    artifact_ids=[
                        package_build_log_1.id,
                        package_build_log_2.id,
                    ],
                ),
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Log filters for Django loggers."""

import logging
from typing import override

from django.db.utils import OperationalError


class SuppressDatabaseErrors(logging.Filter):
    """Suppresses database errors from the logs."""

    @override
    def filter(self, record: logging.LogRecord) -> bool:
        """Filter out database errors from the logs."""
        if (
            record.exc_info is not None
            and record.exc_info[0] is not None
            and issubclass(record.exc_info[0], OperationalError)
        ):
            return False
        return True



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for S3FileBackend."""

import datetime as dt
from io import BytesIO
from typing import ClassVar, override
from urllib.parse import parse_qs, urlsplit

from botocore import stub
from django.utils import timezone

from debusine.assets.models import (
    AWSProviderAccountConfiguration,
    AWSProviderAccountCredentials,
    AWSProviderAccountData,
)
from debusine.db.context import context
from debusine.db.models import Asset, File, FileInStore, FileStore
from debusine.db.tests.utils import _calculate_hash_from_data
from debusine.server.file_backend.s3 import S3FileBackend, S3FileBackendEntry
from debusine.test.django import TestCase
from debusine.test.test_utils import data_generator


class S3FileBackendTests(TestCase):
    """Tests for S3FileBackend."""

    provider_account: ClassVar[Asset]
    file_store: ClassVar[FileStore]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        """Initialize file backend to be tested."""
        super().setUpTestData()
        cls.provider_account = (
            cls.playground.create_cloud_provider_account_asset(
                data=AWSProviderAccountData(
                    name="test",
                    configuration=AWSProviderAccountConfiguration(
                        region_name="test-region"
                    ),
                    credentials=AWSProviderAccountCredentials(
                        access_key_id="access-key",
                        secret_access_key="secret-key",
                    ),
                )
            )
        )
        cls.file_store = FileStore.objects.create(
            name="s3",
            backend=FileStore.BackendChoices.S3,
            configuration={"bucket_name": "test-bucket"},
            provider_account=cls.provider_account,
        )

    def test_no_provider_account(self) -> None:
        """`S3FileBackend` requires a cloud provider account."""
        self.file_store.provider_account = None

        with self.assertRaisesRegex(
            RuntimeError, r"^S3 file backend requires a cloud provider account$"
        ):
            self.file_store.get_backend_object()

    def test_wrong_provider_account(self) -> None:
        """`S3FileBackend` requires an AWS cloud provider account."""
        assert self.file_store.provider_account is not None
        with context.disable_permission_checks():
            self.file_store.provider_account.data = {"provider_type": "hetzner"}
            self.file_store.provider_account.save()

        with self.assertRaisesRegex(
            RuntimeError,
            r"^S3 file backend requires an AWS cloud provider account$",
        ):
            self.file_store.get_backend_object()

    def test_attributes(self) -> None:
        """Initializing `S3FileBackend` sets suitable attributes."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        self.assertEqual(backend.db_store, self.file_store)
        self.assertEqual(backend.client.__class__.__name__, "S3")
        provider_account_data = self.provider_account.data_model
        assert isinstance(provider_account_data, AWSProviderAccountData)
        self.assertEqual(
            backend.client.meta.region_name,
            provider_account_data.configuration.region_name,
        )
        # Private attribute not exposed by type annotations.
        signer = getattr(backend.client, "_request_signer")
        self.assertEqual(
            signer._credentials.access_key,
            provider_account_data.credentials.access_key_id,
        )
        self.assertEqual(
            signer._credentials.secret_key,
            provider_account_data.credentials.secret_access_key,
        )

    def test_endpoint_url(self) -> None:
        """The provider account can set an explicit endpoint URL."""
        assert self.file_store.provider_account is not None
        with context.disable_permission_checks():
            provider_account_data = self.file_store.provider_account.data_model
            assert isinstance(provider_account_data, AWSProviderAccountData)
            provider_account_data.configuration.s3_endpoint_url = (
                "https://s3.eu-west-2.amazonaws.com/"
            )
            self.file_store.provider_account.data = (
                provider_account_data.model_dump(exclude_unset=True)
            )
            self.file_store.provider_account.save()

        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        self.assertEqual(
            backend.client.meta.endpoint_url,
            provider_account_data.configuration.s3_endpoint_url,
        )

    def test_entry_str(self) -> None:
        """`S3FileBackendEntry.__str__` returns the identifier."""
        backend = self.file_store.get_backend_object()
        file = self.playground.create_file()
        entry = backend.get_entry(file)
        self.assertEqual(str(entry), f"{file.sha256.hex()}-{file.size}")

    def test_get_local_path(self) -> None:
        """`S3FileBackend.get_local_path` returns None."""
        backend = self.file_store.get_backend_object()
        file = self.playground.create_file()
        FileInStore.objects.create(store=self.file_store, file=file)
        self.assertIsNone(backend.get_local_path(file))

    def test_get_temporary_local_path(self) -> None:
        """`S3FileBackendEntry.get_temporary_local_path` yields a local path."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = File.from_local_path(temp_file_path)
        identifier = f"{file.sha256.hex()}-{len(contents)}"
        stubber = stub.Stubber(backend.client)
        stubber.add_response(
            "head_object",
            {"ContentLength": len(contents)},
            {"Bucket": "test-bucket", "Key": identifier},
        )
        stubber.add_response(
            "get_object",
            {"Body": BytesIO(contents)},
            {"Bucket": "test-bucket", "Key": identifier},
        )

        with stubber, backend.get_temporary_local_path(file) as temp_path:
            self.assertEqual(temp_path.read_bytes(), contents)

        stubber.assert_no_pending_responses()

    def test_get_url(self) -> None:
        """`S3FileBackend.get_url` returns a presigned URL."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        file = self.playground.create_file()
        FileInStore.objects.create(store=self.file_store, file=file)

        url = backend.get_url(file)
        assert url is not None
        split_url = urlsplit(url)

        self.assertEqual(split_url.scheme, "https")
        self.assertEqual(split_url.netloc, "test-bucket.s3.amazonaws.com")
        self.assertEqual(split_url.path, f"/{file.sha256.hex()}-{file.size}")
        query = parse_qs(split_url.query)
        self.assertRegex(
            query["X-Amz-Credential"][-1], r"^access-key/.*/test-region/.*"
        )
        self.assertEqual(query["X-Amz-Expires"], ["60"])
        self.assertIn("X-Amz-Signature", query)
        self.assertNotIn("response-content-type", query)

    def test_get_url_content_type(self) -> None:
        """`S3FileBackend.get_url` can override the Content-Type."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        file = self.playground.create_file()
        FileInStore.objects.create(store=self.file_store, file=file)

        url = backend.get_url(file, content_type="text/plain")
        assert url is not None
        split_url = urlsplit(url)

        self.assertEqual(split_url.scheme, "https")
        self.assertEqual(split_url.netloc, "test-bucket.s3.amazonaws.com")
        self.assertEqual(split_url.path, f"/{file.sha256.hex()}-{file.size}")
        query = parse_qs(split_url.query)
        self.assertRegex(
            query["X-Amz-Credential"][-1], r"^access-key/.*/test-region/.*"
        )
        self.assertEqual(query["X-Amz-Expires"], ["60"])
        self.assertIn("X-Amz-Signature", query)
        self.assertEqual(query["response-content-type"], ["text/plain"])

    def test_entry_add(self) -> None:
        """`S3FileBackendEntry.add` uploads the file."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = File.from_local_path(temp_file_path)
        entry = backend.get_entry(file)
        identifier = (
            f"{_calculate_hash_from_data(contents).hex()}-{len(contents)}"
        )
        stubber = stub.Stubber(backend.client)
        stubber.add_response(
            "put_object",
            {},
            {
                "Bucket": "test-bucket",
                "Key": identifier,
                "Body": stub.ANY,
                "ChecksumAlgorithm": "SHA256",
            },
        )

        with stubber:
            entry.add(temp_file_path)

        stubber.assert_no_pending_responses()

    def test_entry_add_storage_class(self) -> None:
        """`S3FileBackendEntry.add` can specify a different storage class."""
        self.file_store.configuration["storage_class"] = "EXPRESS_ONEZONE"
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = File.from_local_path(temp_file_path)
        entry = backend.get_entry(file)
        identifier = (
            f"{_calculate_hash_from_data(contents).hex()}-{len(contents)}"
        )
        stubber = stub.Stubber(backend.client)
        stubber.add_response(
            "put_object",
            {},
            {
                "Bucket": "test-bucket",
                "Key": identifier,
                "Body": stub.ANY,
                "ChecksumAlgorithm": "SHA256",
                "StorageClass": "EXPRESS_ONEZONE",
            },
        )

        with stubber:
            entry.add(temp_file_path)

        stubber.assert_no_pending_responses()

    def test_entry_add_hetzner(self) -> None:
        """Skip ChecksumAlgorithm when uploading to Hetzner."""
        assert self.file_store.provider_account is not None
        with context.disable_permission_checks():
            provider_account_data = self.file_store.provider_account.data_model
            assert isinstance(provider_account_data, AWSProviderAccountData)
            provider_account_data.configuration.region_name = None
            provider_account_data.configuration.s3_endpoint_url = (
                "https://hel1.your-objectstorage.com/"
            )
            self.file_store.provider_account.data = (
                provider_account_data.model_dump(exclude_unset=True)
            )
            self.file_store.provider_account.save()

        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = File.from_local_path(temp_file_path)
        entry = backend.get_entry(file)
        identifier = (
            f"{_calculate_hash_from_data(contents).hex()}-{len(contents)}"
        )
        stubber = stub.Stubber(backend.client)
        stubber.add_response(
            "put_object",
            {},
            {
                "Bucket": "test-bucket",
                "Key": identifier,
                "Body": stub.ANY,
                "ChecksumAlgorithm": "SHA256",
            },
        )

        with stubber:
            entry.add(temp_file_path)

        stubber.assert_no_pending_responses()

    def test_entry_same_contents_match(self) -> None:
        """`S3FileBackendEntry.same_contents` accepts matching contents."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = File.from_local_path(temp_file_path)
        entry = backend.get_entry(file)
        identifier = f"{file.sha256.hex()}-{len(contents)}"
        stubber = stub.Stubber(backend.client)
        stubber.add_response(
            "head_object",
            {"ContentLength": len(contents)},
            {"Bucket": "test-bucket", "Key": identifier},
        )
        stubber.add_response(
            "get_object",
            {"Body": BytesIO(contents)},
            {"Bucket": "test-bucket", "Key": identifier},
        )

        with stubber:
            self.assertTrue(entry.same_contents(temp_file_path))

        stubber.assert_no_pending_responses()

    def test_entry_same_contents_mismatch(self) -> None:
        """`S3FileBackendEntry.same_contents` rejects mismatching contents."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = File.from_local_path(temp_file_path)
        entry = backend.get_entry(file)
        identifier = f"{file.sha256.hex()}-{len(contents)}"
        stubber = stub.Stubber(backend.client)
        stubber.add_response(
            "head_object",
            {"ContentLength": len(contents)},
            {"Bucket": "test-bucket", "Key": identifier},
        )
        stubber.add_response(
            "get_object",
            {"Body": BytesIO(contents.lower())},
            {"Bucket": "test-bucket", "Key": identifier},
        )

        with stubber:
            self.assertFalse(entry.same_contents(temp_file_path))

        stubber.assert_no_pending_responses()

    def test_entry_remove(self) -> None:
        """`S3FileBackendEntry.remove` deletes the file."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = File.from_local_path(temp_file_path)
        entry = backend.get_entry(file)
        identifier = f"{file.sha256.hex()}-{len(contents)}"
        stubber = stub.Stubber(backend.client)
        stubber.add_response(
            "delete_object", {}, {"Bucket": "test-bucket", "Key": identifier}
        )

        with stubber:
            entry.remove()

        stubber.assert_no_pending_responses()

    def test_list_entries(self) -> None:
        """`S3FileBackend.list_entries` returns all entries in the backend."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        data_gen = data_generator(16)
        files = [
            self.playground.create_file(contents=next(data_gen))
            for _ in range(10)
        ]
        expected_entries: list[S3FileBackendEntry] = []
        for file in files:
            hash_hex = file.sha256.hex()
            expected_entries.append(
                S3FileBackendEntry(
                    backend=backend, identifier=f"{hash_hex}-{file.size}"
                )
            )
        now = timezone.now()
        stubber = stub.Stubber(backend.client)
        stubber.add_response(
            "list_objects_v2",
            {
                "Contents": [
                    {
                        "Key": f"{file.sha256.hex()}-{file.size}",
                        "LastModified": now - dt.timedelta(days=10 - i),
                    }
                    for i, file in enumerate(files)
                ]
            },
            {"Bucket": "test-bucket"},
        )

        with stubber:
            entries = list(backend.list_entries())
            self.assertCountEqual(entries, expected_entries)
            self.assertEqual(
                {str(entry): entry.get_mtime() for entry in entries},
                {
                    f"{file.sha256.hex()}-{file.size}": now
                    - dt.timedelta(days=10 - i)
                    for i, file in enumerate(files)
                },
            )

        stubber.assert_no_pending_responses()

    def test_list_entries_empty(self) -> None:
        """`S3FileBackend.list_entries` handles an empty backend."""
        backend = self.file_store.get_backend_object()
        assert isinstance(backend, S3FileBackend)
        stubber = stub.Stubber(backend.client)
        stubber.add_response("list_objects_v2", {}, {"Bucket": "test-bucket"})

        with stubber:
            self.assertCountEqual(backend.list_entries(), [])

        stubber.assert_no_pending_responses()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Consumers for the server application."""

import asyncio
import datetime as dt
import json
import logging
from typing import Any, TYPE_CHECKING, cast, override
from urllib.parse import parse_qsl

from asgiref.typing import WebSocketScope
from channels.db import database_sync_to_async
from channels.generic.websocket import AsyncWebsocketConsumer
from prometheus_client import Counter, Gauge
from rest_framework.fields import DateTimeField
from rest_framework.permissions import BasePermission
from rest_framework.renderers import JSONRenderer

from debusine.db.models import Token, WorkRequest, Worker, Workspace
from debusine.server.serializers import OnWorkRequestCompleted
from debusine.server.views.rest import (
    IsUserTokenAuthenticated,
    IsWorkerTokenAuthenticated,
)
from debusine.tasks.models import WorkerType

if TYPE_CHECKING:
    from channels.consumer import _ChannelScope
else:
    _ChannelScope = WebSocketScope

logger = logging.getLogger(__name__)


class ChannelScopeWithToken(_ChannelScope):
    """
    A channel websocket scope with an additional token.

    The token is stored by ``TokenLastSeenAtMiddlewareChannels``.
    """

    token: Token | None


class RequestAdaptor:
    """Minimal class used when helper classes expect a Request."""

    def __init__(self, scope: ChannelScopeWithToken) -> None:
        """Initialize RequestAdaptor."""
        self.auth = scope["token"]


websocket_connect_requests = Counter(
    "debusine_websocket_connect_requests",
    "Incoming websocket connections",
    labelnames=("view",),
)
active_websocket_connections = Gauge(
    "debusine_active_websocket_connections",
    "Active websocket connections",
    labelnames=("view",),
    multiprocess_mode="livesum",
)


class DebusineAsyncWebsocketConsumer(AsyncWebsocketConsumer):
    """Asynchronous websocket consumer with Debusine customizations."""

    # In theory we could figure this out from the URL routing process, but
    # websocket consumers don't have access to a ResolverMatch object, so
    # we'd have to rerun the routing process and that seems unnecessarily
    # heavyweight for metrics.  Just require consumers to declare their view
    # name, which should match `debusine.server.routing`.
    view_name: str

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize metrics."""
        super().__init_subclass__(**kwargs)

        active_websocket_connections.labels(view=cls.view_name).set(0)

    @override
    async def connect(self) -> None:
        """Record the connection."""
        await super().connect()

        websocket_connect_requests.labels(view=self.view_name).inc()
        active_websocket_connections.labels(view=self.view_name).inc()

    def get_header_value(self, header_field_name: str) -> str | None:
        """
        Return the value of header_field_name from self.scope['headers'].

        :param header_field_name: case-insensitive, utf-8 encoded.
        :return: None if header_field_name is not found in
          self.scope['headers'] or the header's content.
        """
        encoded_header_field_name = header_field_name.lower().encode('utf-8')

        for name, value in self.scope['headers']:
            if name == encoded_header_field_name:
                assert isinstance(value, bytes)
                return value.decode('utf-8')

        return None

    async def reject_connection(
        self, reason: str, *, reason_code: str | None = None
    ) -> None:
        """Send JSON rejecting the connection and logs it."""
        msg = {
            'type': 'connection_rejected',
            'reason': reason,
        }
        if reason_code is not None:
            msg["reason_code"] = reason_code

        assert isinstance(self, AsyncWebsocketConsumer)
        await self.send(text_data=json.dumps(msg), close=True)

        logger.info("Consumer rejected. %s", reason)

    async def has_permission(
        self, permission_class: type[BasePermission]
    ) -> bool:
        """Return bool checking if the request has permission to connect."""
        request_adaptor = RequestAdaptor(
            cast(ChannelScopeWithToken, self.scope)
        )

        has_permission = await database_sync_to_async(
            permission_class().has_permission
        )(
            # These arguments don't have the correct types, but we know that
            # our `has_permission` implementations only care about
            # `request.auth`.
            request_adaptor,  # type: ignore[arg-type]
            None,  # type: ignore[arg-type]
        )
        assert isinstance(has_permission, bool)
        return has_permission

    @override
    async def disconnect(self, close_code: Any) -> None:
        """Record the disconnection."""
        active_websocket_connections.labels(view=self.view_name).dec()

        await super().disconnect(close_code)


class WorkerConsumer(DebusineAsyncWebsocketConsumer):
    """
    Implement server-side of a Worker.

    After the client worker connects to the server: WorkerConsumer
    requests information, send tasks, etc.
    """

    view_name = "ws-connect"

    # Call self._send_dynamic_metadata_request() every
    # REQUEST_DYNAMIC_METADATA_SECONDS
    REQUEST_DYNAMIC_METADATA_SECONDS = 3600

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize WorkerConsumer member variables."""
        super().__init__(*args, **kwargs)
        self._worker: Worker | None = None
        self._request_dynamic_metadata_task: asyncio.Task[None] | None = None

    async def _channel_layer_group_add(
        self, token_hash: str, channel_name: str
    ) -> None:
        await self.channel_layer.group_add(token_hash, channel_name)

    @override
    async def connect(self) -> None:
        """Worker client is connecting."""
        await super().connect()

        if not await self.has_permission(IsWorkerTokenAuthenticated):
            await self.reject_connection(
                "No token header, token not associated to a worker "
                "or not enabled",
                reason_code="TOKEN_DISABLED",
            )
            return

        token_header: str | None = self.get_header_value("token")
        token_hash: str | None = (
            Token._generate_hash(token_header) if token_header else None
        )

        self._worker = await Worker.objects.select_related("token").aget(
            token__hash=token_hash
        )
        if self._worker.worker_type not in (
            WorkerType.EXTERNAL,
            WorkerType.SIGNING,
        ):
            await self.reject_connection(
                'Unexpected worker type, only EXTERNAL and SIGNING supported'
            )
            return

        # Must be true, since we already checked IsWorkerTokenAuthenticated.
        assert self._worker is not None
        assert self._worker.token is not None

        try:
            await self._channel_layer_group_add(
                self._worker.token.hash, self.channel_name
            )
        except Exception as exc:
            logger.error(  # noqa: G200
                'Error adding worker to group (Redis): %s', exc
            )
            await self.reject_connection('Service unavailable')
            return

        logger.info("Worker connected: %s", self._worker.name)

        await self.send(
            text_data=json.dumps(
                {"type": "websocket.send", "text": "connected"}
            )
        )

        # Detect if we have a work request already affected. If yes, then
        # we trigger a "work_request_available" message at the end so that
        # the worker has a chance to pick up that work request in case
        # it got restarted.
        worker_is_busy = await database_sync_to_async(self._worker.is_busy)()

        # Both the "dynamic_metadata_refresher" and the "mark_connected" methods
        # ultimately end up with an update of the Worker model which triggers a
        # scheduler run that can trigger a "work_request_available" message
        # on its own.
        self._request_dynamic_metadata_task = asyncio.create_task(
            self._dynamic_metadata_refresher(),
            name='dynamic_metadata_refresher',
        )

        await database_sync_to_async(self._worker.mark_connected)()

        if worker_is_busy:
            await self._send_work_request_available()

    async def _send_work_request_available(self) -> None:
        await self.send(
            text_data=json.dumps(
                {
                    "type": "websocket.send",
                    "text": "work_request_available",
                }
            )
        )

    async def _dynamic_metadata_refresher(self) -> None:
        while True:
            await self._send_dynamic_metadata_request()
            await asyncio.sleep(self.REQUEST_DYNAMIC_METADATA_SECONDS)

    async def _send_dynamic_metadata_request(self) -> None:
        await self.send(
            text_data=json.dumps(
                {
                    "type": "websocket.send",
                    "text": "request_dynamic_metadata",
                }
            )
        )

    @override
    async def disconnect(self, close_code: Any) -> None:
        """Worker has disconnected. Cancel tasks, mark as disconnect, etc."""
        await super().disconnect(close_code)

        if self._request_dynamic_metadata_task:
            self._request_dynamic_metadata_task.cancel()

        if self._worker:
            assert self._worker.token is not None

            await self.channel_layer.group_discard(
                self._worker.token.hash, self.channel_name
            )

            await self._worker.arefresh_from_db()
            await database_sync_to_async(self._worker.mark_disconnected)()
            logger.info(
                "Worker disconnected: %s (code: %s)",
                self._worker.name,
                close_code,
            )

    async def worker_disabled(
        self,
        event: dict[str, Any],  # noqa: ARG002, U100
    ) -> None:
        """Worker has been disabled. Send a connection_closed msg."""
        assert self._worker is not None
        assert self._worker.token is not None

        logger.info("Worker %s disabled", self._worker.name)

        msg = {
            "type": "connection_closed",
            "reason": 'Token has been disabled '
            f'(token hash: "{self._worker.token.hash}")',
            "reason_code": "TOKEN_DISABLED",
        }
        await self.send(text_data=json.dumps(msg), close=True)

    async def work_request_assigned(
        self,
        event: dict[str, Any],  # noqa: ARG002, U100
    ) -> None:
        """Work Request has been assigned to the worker. Send channel msg."""
        await self._send_work_request_available()


class OnWorkRequestCompletedConsumer(DebusineAsyncWebsocketConsumer):
    """Send work_request_completed notifications."""

    view_name = "ws-work-request-on-completed"

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize OnWorkRequestCompletedConsumer."""
        super().__init__(*args, **kwargs)

        # If _workspaces is not None: set of the workspaces that are being
        # monitored, otherwise all the workspaces that the client has access to
        self._workspaces_only: set[int] | None = None

        self._query_string: dict[str, str] = {}

        # RedisPubSubChannelLayer in channels_redis 4.0.0 has a bug in
        # group_discard where it fails if group_add hadn't been called.
        # This was fixed in 4.1.0, but work around it.
        self._channel_added_to_group: bool = False

    async def _get_workspaces_from_query_params(self) -> set[int] | None:
        workspace_names_str = self._query_string.get("workspaces")

        if workspace_names_str is None:
            return None

        workspace_names = workspace_names_str.split(",")

        def get_workspace_ids(workspace_names: list[str]) -> set[int]:
            # TODO: check that the current user has access to each
            # workspace (need implementing #80)
            names_to_ids = {
                workspace.name: workspace.id
                for workspace in Workspace.objects.filter(
                    name__in=workspace_names
                )
            }
            for workspace_name in workspace_names:
                if workspace_name not in names_to_ids:
                    raise Workspace.DoesNotExist(
                        f'Workspace "{workspace_name}" not found'
                    )
            return set(names_to_ids.values())

        workspace_ids = await database_sync_to_async(get_workspace_ids)(
            workspace_names
        )
        assert isinstance(workspace_ids, set)

        return workspace_ids

    @override
    async def connect(self) -> None:
        """Client is connecting."""
        self._query_string = dict(
            parse_qsl(self.scope["query_string"].decode("utf-8"))
        )

        await super().connect()

        if not await self.has_permission(IsUserTokenAuthenticated):
            await self.reject_connection(
                "No token header, token not associated to a user or not enabled"
            )
            return

        try:
            self._workspaces_only = (
                await self._get_workspaces_from_query_params()
            )
        except Workspace.DoesNotExist as exc:
            await self.reject_connection(str(exc))
            return

        await self.channel_layer.group_add(
            "work_request_completed", self.channel_name
        )
        self._channel_added_to_group = True

        await self.send(
            text_data=json.dumps(
                {"type": "websocket.send", "text": "connected"}
            )
        )

        await self._send_pending_work_requests_since_last_completed()

    @override
    async def disconnect(self, close_code: Any) -> None:  # noqa: U100
        """Client is disconnecting, clean-up."""
        await super().disconnect(close_code)

        if self._channel_added_to_group:
            await self.channel_layer.group_discard(
                "work_request_completed", self.channel_name
            )
        self._channel_added_to_group = False

    async def _get_pending_work_requests(
        self, completed_at_since: dt.datetime
    ) -> list[WorkRequest]:
        """
        Return WorkRequests that completed_at on or after completed_at_since.

        If there is one WorkRequest that completed_at == completed_at_since:
        do not include this one (the client was already notified).

        If more than one WorkRequest completed_at == completed_at_since:
        include both of them. The client might have not processed all of them.
        """
        on_last_completed_count = await WorkRequest.objects.filter(
            completed_at=completed_at_since
        ).acount()

        if on_last_completed_count == 1:
            filter_kwargs = {"completed_at__gt": completed_at_since}
        else:
            filter_kwargs = {"completed_at__gte": completed_at_since}

        return [
            wr
            async for wr in WorkRequest.objects.filter(
                **filter_kwargs
            ).order_by("completed_at", "id")
        ]

    async def _send_pending_work_requests_since_last_completed(self) -> None:
        completed_at_since = self._query_string.get("completed_at_since")

        if completed_at_since is None:
            return

        for work_request in await self._get_pending_work_requests(
            DateTimeField().to_internal_value(completed_at_since)
        ):
            assert work_request.completed_at is not None
            await self._send_work_request_completed(
                work_request.id,
                work_request.workspace_id,
                work_request.completed_at.isoformat(),
                work_request.result,
            )

    async def work_request_completed(self, event: dict[str, Any]) -> None:
        """Work Request has completed."""
        work_request_id = event["work_request_id"]
        work_space_id = event["workspace_id"]
        completed_at = event["completed_at"]
        result = event["result"]

        await self._send_work_request_completed(
            work_request_id, work_space_id, completed_at, result
        )

    async def _send_work_request_completed(
        self,
        work_request_id: int,
        workspace_id: int,
        completed_at: str,
        result: str,
    ) -> None:
        if (
            self._workspaces_only is not None
            and workspace_id not in self._workspaces_only
        ):
            return

        serializer = OnWorkRequestCompleted(
            data={
                "type": "websocket.send",
                "text": "work_request_completed",
                "work_request_id": work_request_id,
                "completed_at": dt.datetime.fromisoformat(completed_at),
                "result": result,
            }
        )

        serializer.is_valid(raise_exception=True)

        await self.send(
            text_data=JSONRenderer().render(serializer.data).decode("utf-8")
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command signon_sync."""

from typing import ClassVar, override
from unittest import mock

import gitlab
import gitlab.v4.objects
from django.test import override_settings

from debusine.db.models import Group
from debusine.db.models.auth import Identity
from debusine.db.playground import scenarios
from debusine.django.management.tests import call_command
from debusine.server.signon import providers
from debusine.test.django import TestCase


class SignonSyncCommandTests(TestCase):
    """Tests for signon_sync command."""

    scenario = scenarios.DefaultContext()
    group: ClassVar[Group]
    identity: ClassVar[Identity]

    @override
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.group = cls.playground.create_group(name="group")
        cls.identity = Identity.objects.create(
            user=cls.scenario.user, issuer="salsa", subject="123"
        )

    def mock_gitlab(self) -> gitlab.Gitlab:
        """Return a mock gitlab object."""
        gl = mock.MagicMock(spec=gitlab.Gitlab)
        gl.groups = mock.MagicMock(spec=gitlab.v4.objects.GroupManager)
        gl.groups.get = mock.MagicMock()
        return gl

    def mock_gitlab_group(self) -> gitlab.v4.objects.Group:
        """Return a mock gitlab group."""
        group = mock.MagicMock(spec=gitlab.v4.objects.Group)
        group.members_all = mock.MagicMock(
            spec=gitlab.v4.objects.GroupMemberAllManager
        )
        group.members_all.list = mock.MagicMock()
        return group

    def mock_gitlab_user(self, id: int | None = None) -> gitlab.v4.objects.User:  # noqa: A002
        """Return a mock gitlab user."""
        user = mock.MagicMock(spec=gitlab.v4.objects.User)
        user.id = id if id is not None else int(self.identity.subject)
        return user

    @override
    def setUp(self) -> None:
        super().setUp()
        self.provider = providers.GitlabProvider(
            name="salsa",
            label="Salsa",
            url="https://salsa.debian.org",
            client_id="client_id",
            client_secret="secret",
            options={
                "add_to_group": {"remote": str(self.group)},
                "sync_token": "TOKEN123",
            },
        )
        self.gl = self.mock_gitlab()
        self.enterContext(mock.patch("gitlab.Gitlab", return_value=self.gl))

    def test_sync_unsupported_provider(self) -> None:
        self.enterContext(
            override_settings(
                SIGNON_PROVIDERS=[
                    providers.Provider("unsupported", "Unsupported")
                ]
            )
        )
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(), ["unsupported: provider unsupported for sync"]
        )
        self.assertEqual(retval, 0)

    def test_sync_gitlab_no_add_to_group(self) -> None:
        del self.provider.options["add_to_group"]
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(),
            ["salsa: provider without add_to_group: skipping sync"],
        )
        self.assertEqual(retval, 0)

    def test_sync_gitlab_no_sync_token(self) -> None:
        del self.provider.options["sync_token"]
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(),
            [
                "salsa: Provider salsa has no 'sync_token'"
                " configured in its options"
            ],
        )
        self.assertEqual(retval, 0)

    def test_sync_gitlab_nm_remote(self) -> None:
        self.provider.options["add_to_group"] = {
            "nm:unsupported": str(self.group)
        }
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(stdout, "")
        self.assertEqual(
            stderr.splitlines(),
            ["nm:unsupported: sync with nm.debian.org is not yet implemented"],
        )
        self.assertEqual(retval, 0)

    def test_sync_gitlab(self) -> None:
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(
            stdout.splitlines(),
            [
                "remote → debusine/group: 0 valid members, 0 members to add,"
                " 0 members to remove, 0 users not found in Debusine"
            ],
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_sync_gitlab_add_user_dry_run(self) -> None:
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        gl_user = self.mock_gitlab_user()
        gl_group = self.mock_gitlab_group()
        gl_group.members_all.list.return_value = [gl_user]
        self.gl.groups.get.return_value = gl_group
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(
            stdout.splitlines(),
            [
                "remote → debusine/group: 0 valid members, 1 members to add,"
                " 0 members to remove, 0 users not found in Debusine",
                "playground should be added to debusine/group",
            ],
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        self.assertQuerySetEqual(self.group.users.all(), [])

    def test_sync_gitlab_add_user(self) -> None:
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        gl_user = self.mock_gitlab_user()
        gl_group = self.mock_gitlab_group()
        gl_group.members_all.list.return_value = [gl_user]
        self.gl.groups.get.return_value = gl_group
        stdout, stderr, retval = call_command("signon_sync", "--fix")
        self.assertEqual(
            stdout.splitlines(),
            [
                "remote → debusine/group: 0 valid members, 1 members to add,"
                " 0 members to remove, 0 users not found in Debusine",
                "playground added to debusine/group",
            ],
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        self.assertQuerySetEqual(self.group.users.all(), [self.scenario.user])

    def test_sync_gitlab_remove_user_dry_run(self) -> None:
        self.playground.add_user(self.group, self.scenario.user)
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        gl_group = self.mock_gitlab_group()
        gl_group.members_all.list.return_value = []
        self.gl.groups.get.return_value = gl_group
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(
            stdout.splitlines(),
            [
                "remote → debusine/group: 0 valid members, 0 members to add,"
                " 1 members to remove, 0 users not found in Debusine",
                "playground should be removed from debusine/group",
            ],
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        self.assertQuerySetEqual(self.group.users.all(), [self.scenario.user])

    def test_sync_gitlab_remove_user(self) -> None:
        self.playground.add_user(self.group, self.scenario.user)
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        gl_group = self.mock_gitlab_group()
        gl_group.members_all.list.return_value = []
        self.gl.groups.get.return_value = gl_group
        stdout, stderr, retval = call_command("signon_sync", "--fix")
        self.assertEqual(
            stdout.splitlines(),
            [
                "remote → debusine/group: 0 valid members, 0 members to add,"
                " 1 members to remove, 0 users not found in Debusine",
                "playground removed from debusine/group",
            ],
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

        self.assertQuerySetEqual(self.group.users.all(), [])

    def test_sync_gitlab_unknown_user(self) -> None:
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        gl_user = self.mock_gitlab_user(id=12345)
        gl_group = self.mock_gitlab_group()
        gl_group.members_all.list.return_value = [gl_user]
        self.gl.groups.get.return_value = gl_group
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(
            stdout.splitlines(),
            [
                "remote → debusine/group: 0 valid members, 0 members to add,"
                " 0 members to remove, 1 users not found in Debusine"
            ],
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_sync_gitlab_valid_user(self) -> None:
        self.playground.add_user(self.group, self.scenario.user)
        self.enterContext(override_settings(SIGNON_PROVIDERS=[self.provider]))
        gl_user = self.mock_gitlab_user()
        gl_group = self.mock_gitlab_group()
        gl_group.members_all.list.return_value = [gl_user]
        self.gl.groups.get.return_value = gl_group
        stdout, stderr, retval = call_command("signon_sync")
        self.assertEqual(
            stdout.splitlines(),
            [
                "remote → debusine/group: 1 valid members, 0 members to add,"
                " 0 members to remove, 0 users not found in Debusine"
            ],
        )
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: base infrastructure."""

import copy
import logging
from collections.abc import Callable
from typing import (
    Any,
    Protocol,
    Self,
    TYPE_CHECKING,
    TypeAlias,
    override,
    runtime_checkable,
)

from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Model, QuerySet
from django.http import Http404, HttpRequest, HttpResponseBase
from rest_framework import status, viewsets
from rest_framework.filters import BaseFilterBackend
from rest_framework.generics import (
    CreateAPIView,
    DestroyAPIView,
    GenericAPIView,
    ListAPIView,
    RetrieveAPIView,
    UpdateAPIView,
)
from rest_framework.request import Request
from rest_framework.response import Response
from rest_framework.serializers import BaseSerializer
from rest_framework.views import APIView

from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import Workspace
from debusine.db.models.permissions import (
    anonymous_access,
    format_permission_check_error,
)
from debusine.db.permissioncontext import PermissionContext, PermissionUser
from debusine.server.exceptions import (
    DebusineAPIException,
    raise_workspace_not_found,
)
from debusine.server.views.rest import ProblemResponse

if TYPE_CHECKING:
    from rest_framework.permissions import _PermissionClass

    _PermissionClass  # fake usage for vulture
    CreateAPIViewBase: TypeAlias = CreateAPIView
    DestroyAPIViewBase: TypeAlias = DestroyAPIView
    GenericAPIViewBase: TypeAlias = GenericAPIView
    ListAPIViewBase: TypeAlias = ListAPIView
    RetrieveAPIViewBase: TypeAlias = RetrieveAPIView
    UpdateAPIViewBase: TypeAlias = UpdateAPIView
    GenericViewSetBase: TypeAlias = viewsets.GenericViewSet
else:
    # REST framework's generic views don't support generic types at run-time
    # yet.
    from debusine.utils.typing_utils import patch_runtime_generic_support

    CreateAPIViewBase = patch_runtime_generic_support(CreateAPIView)
    DestroyAPIViewBase = patch_runtime_generic_support(DestroyAPIView)
    GenericAPIViewBase = patch_runtime_generic_support(GenericAPIView)
    GenericViewSetBase = patch_runtime_generic_support(viewsets.GenericViewSet)
    ListAPIViewBase = patch_runtime_generic_support(ListAPIView)
    RetrieveAPIViewBase = patch_runtime_generic_support(RetrieveAPIView)
    UpdateAPIViewBase = patch_runtime_generic_support(UpdateAPIView)


class BaseAPIView(APIView):
    """Common base for API views."""

    permissions_checked: bool = False

    def enforce(self, predicate: Callable[[PermissionContext], bool]) -> None:
        """Enforce a permission predicate."""
        self.permissions_checked = True
        if predicate(context.pc):
            return

        raise DebusineAPIException(
            title=format_permission_check_error(predicate, context.pc),
            status_code=status.HTTP_403_FORBIDDEN,
        )

    @override
    def dispatch(
        self,
        request: HttpRequest,
        *args: Any,
        **kwargs: Any,
    ) -> HttpResponseBase:
        response = super().dispatch(request, *args, **kwargs)
        if 200 <= response.status_code < 300 and not self.permissions_checked:
            raise AssertionError("View does not enforce permission checks.")
        return response

    def set_current_workspace(
        self,
        workspace: str | Workspace,
    ) -> None:
        """Set the current workspace in context."""
        if isinstance(workspace, str):
            try:
                workspace = Workspace.objects.get_for_context(name=workspace)
            except Workspace.DoesNotExist:
                raise_workspace_not_found(workspace)
        try:
            workspace.set_current()
        except ContextConsistencyError:
            # Turn exception in a 404 response.
            # 404 is used instead of 403 as an attempt to prevent leaking which
            # private workspace exists that the user cannot see
            logging.debug("permission denied on %s reported as 404", workspace)
            raise_workspace_not_found(workspace)


class GetOrCreateAPIView[M: Model](CreateAPIViewBase[M], BaseAPIView):
    """Class to create a new object (return 201 created) or return 200."""

    @override
    def get_object(self) -> M:
        """Override to return more API-friendly errors."""
        try:
            return super().get_object()
        except Http404 as exc:
            raise DebusineAPIException(
                title=str(exc), status_code=status.HTTP_404_NOT_FOUND
            )

    def get_existing_object(self, serializer: BaseSerializer[M]) -> M:
        """
        If there is an existing object matching Serializer, return it.

        If not, raise ObjectDoesNotExist.
        """
        return self.get_queryset().get(**serializer.validated_data)

    @override
    def create(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        """
        Return the object with HTTP 201 (created) or HTTP 200 (already existed).

        It is a modification of CreateAPIView.create().

        The default behaviour in rest_framework when an object cannot be
        inserted because it already exists is to return HTTP 400. In
        GetOrCreateApiView it returns HTTP 200 and the client knows that
        the object that wanted to create was already in the database.
        """
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        response_status: int
        try:
            obj = self.get_existing_object(serializer)
            response_status = status.HTTP_200_OK
            pk = obj.pk
        except ObjectDoesNotExist:
            self.perform_create(serializer)
            response_status = status.HTTP_201_CREATED
            assert serializer.instance is not None
            pk = serializer.instance.pk

        headers = self.get_success_headers(serializer.data)

        response_data = copy.deepcopy(serializer.data)
        response_data["id"] = pk

        return Response(response_data, status=response_status, headers=headers)


@runtime_checkable
class SupportsCanDisplay(Protocol):
    """A query set that supports `can_display` filtering."""

    def can_display(self, user: PermissionUser) -> Self:
        """Keep only objects that can be displayed."""


class CanDisplayFilterBackend(BaseFilterBackend):
    """Filter that only allows objects that can be displayed."""

    @override
    def filter_queryset[M: Model, Row = Model](
        self,
        request: Request,  # noqa: ARG002, U100
        queryset: QuerySet[M, Row],
        view: APIView,  # noqa: ARG002, U100
    ) -> QuerySet[M, Row]:
        """Keep only objects that can be displayed."""
        assert isinstance(queryset, SupportsCanDisplay)
        # This may be called in contexts where no user is set: for example,
        # some tasks need to call views using this filter to list artifact
        # relations, in which case the request is authenticated using a
        # worker token and context.user is unset.  The permission_filter
        # decorator means that it's actually OK to call
        # Workspace.objects.can_display with None, but for some reason the
        # type checker doesn't understand that.
        return queryset.can_display(context.pc)  # type: ignore[arg-type]


class Whoami(BaseAPIView):
    """Simple view that returns the authentication status."""

    def get(self, request: Request) -> Response:
        """Return information about the current user."""
        self.enforce(anonymous_access)
        user: str
        if request.user.is_authenticated:
            user = request.user.username
        else:
            user = "<anonymous>"
        work_request: int | None = None
        if context.pc.work_request:
            work_request = context.pc.work_request.id
        return Response(
            {
                "user": user,
                "auth": bool(request.auth),
                "work_request": work_request,
                "worker": bool(context.pc.worker_token),
            }
        )


class SupersededAPIView(APIView):
    """
    Simple view that replaces a removed API endpoint.

    To use it, mount it in a URL conf with optional hints for the replacement::

        path(
            "superseded-endpoint/",
            SupersededAPIView.as_view(),
            {"new_endpoint": "/1.0/new-endpoint/", "client_version": "1.0"},
        )
    """

    @override
    def dispatch(
        self,
        request: HttpRequest,
        *args: Any,
        new_endpoint: str | None = None,
        client_version: str | None = None,
        **kwargs: Any,
    ) -> HttpResponseBase:
        """Configure the instance and dispatch a respnose."""
        self.new_endpoint = new_endpoint
        self.client_version = client_version

        if request.method in ("GET", "POST", "PATCH", "DELETE"):
            return self.error_response(new_endpoint, client_version)
        return super().dispatch(request, *args, **kwargs)

    def error_response(
        self, new_endpoint: str | None = None, client_version: str | None = None
    ) -> HttpResponseBase:
        """Return the view's superseded error."""
        detail = "This endpoint is no longer supported."
        if new_endpoint:
            detail += f" It has been replaced by {new_endpoint}"
            if client_version:
                detail += (
                    f", which is supported by debusine-client (>= "
                    f"{client_version}). "
                    f"Please upgrade your client"
                )
            detail += "."

        return ProblemResponse(title="Superseded API Endpoint", detail=detail)
